// $Id: dcc.cc,v 1.54 2003/01/16 16:11:27 toast Exp $

// C headers
#include <sys/errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <fcntl.h>
#include <time.h>
#include <netdb.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <signal.h>
#include <errno.h>

#include "oomon.h"

// C++ headers
#include <iostream>
#include <fstream>
#include <string>
#include <list>
#include "strtype"
#include "config.h"
#include "log.h"
#include "irc.h"
#include "dcc.h"
#include "util.h"
#include "main.h"
#include "engine.h"
#include "help.h"
#include "trap.h"

int DCC::count = 0;
class DCC* DCC::items[MAX_DCCS];

// Connect(address, port, nick, userhost)
//
// Makes a DCC connection to the given address and port.
//
// Returns 0 if no errors ocurred.
//
int DCC::Connect(const unsigned long address, const int port,
        const std::string & nick, const std::string & userhost)
{
  if (count >= MAX_DCCS)
    return INVALID;

  class DCC *temp;

  temp = new DCC;

  if (temp->PrivConnect(address, port, nick, userhost) == 0) {
    //Status("*", userhost + " has connected");
    //Remote::Status("*", userhost + " has connected");
    items[count++] = temp;
    return 0;
  } else {
    delete temp;
    return INVALID;
  }
}

// Listen(nick, userhost)
//
// Listens for a DCC connection
//
// Returns 0 if no errors ocurred.
//
int DCC::Listen(const std::string & nick, const std::string & userhost)
{
  if (count >= MAX_DCCS)
    return INVALID;

  class DCC *temp;

  temp = new DCC;

  if (temp->PrivListen(nick, userhost) == 0) {
    items[count++] = temp;
    return 0;
  } else {
    delete temp;
    return INVALID;
  }
}

// SetFD(readset)
//
// Do this if you want to be able to receive data :P
//
void DCC::SetFD(fd_set & readset, fd_set & writeset)
{
  class DCC *temp;

  for (int i = 0; i < count; i++) {
    temp = items[i];
    temp->PrivSetFD(readset, writeset);
  }
}

// Process(readset, writeset)
//
// Processes any received data at each DCC connection.
//
// Returns 0 if no errors ocurred.
//
int DCC::Process(const fd_set & readset, const fd_set & writeset)
{
  class DCC *temp;
  int i = 0;
  while (i < count) {
    temp = items[i];
    if (temp->PrivProcess(readset, writeset) != 0) {
      std::string uh = temp->GetUserHost();
      std::string nick = temp->GetHandle();
      bool conn = temp->IsConnected();
      delete temp;
      for (int j = i; j < count-1; j++)
        items[j] = items[j+1];
      count--;
      if (conn) {
        Status("*", uh + " has disconnected");
        Remote::Status("*", uh + " has disconnected");
      } else {
	IRC::Notice(nick, "DCC CHAT connection timed out.");
        Log::Write(uh + " timed out on DCC connect");
      }
    } else
      i++;
  }
  return 0;
}

// Write(text, exception)
//
// Writes data to all DCC connections who don't match the exception
//
void DCC::Write(const std::string & text, const class DCC *exception)
{
  for (int i = 0; i < count; i++)
    if (items[i] != exception)
      items[i]->Write(text);
}


// Status(To, Text, exception)
//
// Writes a status message to all DCC connections who don't match the
// exception or only to the indicated handle.
//
void DCC::Status(const std::string & To, const std::string & Text,
        const class DCC *exception)
{
  if (To == "*")
    Write(Text + "\n", exception);
  else {
    class DCC *temp = FindHandle(To);
    if (temp != NULL)
      temp->Write(Text + "\n");
  }
}


// Status(From, To, Text, exception)
//
// Writes a status message to all DCC connections who don't match the
// exception or only to the indicated handle.
//
void DCC::Status(const std::string & From, const std::string & To,
        const std::string & Text, const class DCC *exception)
{
  if (To == "*")
    Write("(" + From + ") " + Text + "\n", exception);
  else {
    class DCC *temp = FindHandle(To);
    if (temp != NULL)
      temp->Write("(" + From + ") " + Text + "\n");
  }
}


// Write(Text, Flags)
//
// Writes a message to all DCC connections that match the given user
// flags.
//
void DCC::WriteByFlags(const std::string & text, const int flags)
{
  for (int i = 0; i < count; i++)
    items[i]->Write(text, flags);
}


// Chat(From, To, Text, exception)
//
// Writes a chat message to all DCC connections who don't match the
// exception.  Returns false if user doesn't exist.
//
bool DCC::Chat(const std::string & From, const std::string & To, std::string Text,
        const class DCC *exception)
{
  if (Text != "") {
    if (Text[0] != '\001')
      if (To == "*")
        Write("<" + From + "> " + Text + "\n", exception);
      else {
        class DCC *temp = FindHandle(To);
        if (temp != NULL)
          temp->Write("*" + From + "* " + Text + "\n");
        else
          return false;
      }
    else {
      // Need to be ready for CTCP ACTION type stuff here.
      Text.erase((std::string::size_type) 0, 1);
      int i = Text.find('\001');
      if (i != std::string::npos)
        Text = Text.substr(0, i);
      std::string Command = FirstWord(Text);
      if ((Command == "ACTION") && (Text != ""))
        if (To == "*")
          Write("* " + From + " " + Text + "\n", exception);
    }  
  }
  return true;
}


// Who(Output)
//
// Returns a list of connected users
//
void DCC::Who(StrList & Output)
{
  char buff[MAX_BUFF];

  Output.clear();
  if (count > 0) {
    Output.push_back("Client connections:");
    for (int i = 0; i < count; i++)
      if (items[i]->IsConnected()) {
        sprintf(buff, "%u", items[i]->GetIdle());
        Output.push_back(items[i]->GetFlags() + " " + items[i]->GetHandle() +
	  " (" + items[i]->GetUserHost() + ") " + std::string(buff));
      } else {
        Output.push_back(items[i]->GetFlags() + " " + items[i]->GetHandle() +
	  " (" + items[i]->GetUserHost() + ") -connecting-");
      }
  } else
    Output.push_back("No client connections!");
}


// StatsP(Output)
//
// Returns a list of connected users (in a pseudo-"stats p" format)
//
void DCC::StatsP(StrList & Output)
{
  std::string BNick = Config::GetNick();
  int operCount = 0;

  Output.clear();
  for (int i = 0; i < count; i++)
  {
    if (items[i]->IsConnected() && items[i]->IsOper())
    {
      Output.push_back(items[i]->GetHandle() + "@" + BNick
#ifdef SHOW_STATSP_USERHOST
	+ std::string(" (" + items[i]->GetUserHost() + ")") 
#endif
#ifdef SHOW_STATSP_IDLE
	+ std::string(" Idle: " + IntToStr(items[i]->GetIdle()))
#endif
	);
      operCount++;
    }
  }
  Output.push_back(IntToStr(operCount) + " OOMon opers");
}

class DCC *DCC::FindHandle(const std::string & handle)
{
  for (int i = 0; i < count; i++)
    if (Same(handle, items[i]->GetHandle()))
      return items[i];
  return NULL;
}


DCC::DCC()
{
  echoMyChatter = Authed = connected = listening = false;
  plug = 0;
  UserHost = "";
  Flags = 0;
}


DCC::~DCC()
{
  if (connected)
    Disconnect();
}


// PrivConnect(address, port, nick, userhost)
//
// Makes a DCC chat connection. Woohoo!
//
// Returns 0 if no errors ocurred
//
int DCC::PrivConnect(const unsigned long address, const int port,
	const std::string & nick, const std::string & userhost)
{
  struct sockaddr_in	socketname;
  struct sockaddr_in	localaddr;
  struct hostent	*remote_host;
  char			buf [MAX_BUFF];

  #ifdef DCC_DEBUG
    cout << "Creating DCC connection to " << nick << " (" << userhost << ")" << endl;
  #endif

  UserHost = userhost;
  Handle = Nick = nick;

  if ((plug = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
    cerr << "error: can't assign fd for socket" << endl;
    return INVALID;
  } 

  int optval = 1;
  setsockopt(plug, SOL_SOCKET, SO_REUSEADDR, &optval, sizeof(optval));
  setsockopt(plug, SOL_SOCKET, SO_KEEPALIVE, &optval, sizeof(optval));

  (void)bzero((char *)&socketname, sizeof(socketname));

  socketname.sin_addr.s_addr=htonl(address);
  socketname.sin_family = AF_INET;
  socketname.sin_port = htons(port);

  if (-1 == fcntl(plug, F_SETFL, O_NONBLOCK)) {
    cerr << "error: can't set fd non-blocking" << endl;
    return INVALID;
  }
  #ifdef DCC_DEBUG
    cout << "Set socket non-blocking." << endl;
  #endif

  if (connect(plug, (struct sockaddr *) &socketname, sizeof(socketname)) < 0)
    if (errno != EINPROGRESS) {
      cerr << "Error: Connect " << errno << " (" << strerror(errno) << ")" << endl;
      return INVALID;
    }

  //connected = true;
  //Write("OOMon-" + std::string(OOMON_VERSION) + " by Timothy L. Jensen\n");
  //MOTD();

  LastActive = time((time_t *)NULL);

  return 0;
}


// PrivListen(nick, userhost)
//
// Listens for a DCC chat connection. Woohoo!
//
// Returns 0 if no errors ocurred
//
int DCC::PrivListen(const std::string & nick, const std::string & userhost)
{
  struct sockaddr_in	socketname;
  socklen_t		size;

  #ifdef DCC_DEBUG
    cout << "Creating DCC connection to " << nick << " (" << userhost << ")" << endl;
  #endif

  UserHost = userhost;
  Handle = Nick = nick;

  if ((plug = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
    cerr << "error: can't assign fd for socket" << endl;
    return INVALID;
  } 

  int optval = 1;
  setsockopt(plug, SOL_SOCKET, SO_REUSEADDR, &optval, sizeof(optval));
  setsockopt(plug, SOL_SOCKET, SO_KEEPALIVE, &optval, sizeof(optval));

  (void)bzero((char *)&socketname, sizeof(socketname));

  socketname.sin_addr.s_addr=htonl(INADDR_ANY);
  socketname.sin_family = AF_INET;
  socketname.sin_port = 0;

  if (bind(plug, (struct sockaddr *)&socketname, sizeof(socketname)) < 0) {
    cerr << "Error: Bind " << errno << " (" << strerror(errno) << ")" << endl;
    return INVALID;
  }

  if (-1 == fcntl(plug, F_SETFL, O_NONBLOCK)) {
    cerr << "error: can't set fd non-blocking" << endl;
    return INVALID;
  }
  #ifdef DCC_DEBUG
    cout << "Set socket non-blocking." << endl;
  #endif

  if (listen(plug, 1) < 0) {
    cerr << "Error: Listen " << errno << " (" << strerror(errno) << ")" << endl;
    return INVALID;
  }

  size = sizeof(socketname);
  getsockname(plug, (struct sockaddr *)&socketname, &size);

  listening = true;

  IRC::CTCP(nick, "DCC CHAT chat " + ULongToStr(IRC::GetMyAddr()) + " " + IntToStr(ntohs(socketname.sin_port)));

  LastActive = time((time_t *)NULL);

  return 0;
}


// Disconnect()
//
// Terminates the DCC connection. Duh!
//
int DCC::Disconnect()
{
  if (connected) {
    #ifdef DCC_DEBUG
      cout << "Closing DCC connection to " << Handle << " (" << UserHost << ")" << endl;
    #endif
    Log::Write("Closing DCC connection to " + Handle + " (" + UserHost + ")");
    close(plug);
  }
  plug = 0;
  connected = false;
  return 0;
}


// Write(msg)
//
// Sends data to the DCC connection
//
// Returns 0 if no errors ocurred.
//
int DCC::Write(const std::string & msg, const int flags)
{
  if (connected) {
    if (!flags || (flags & Flags)) {
      #ifdef DCC_DEBUG
        cout << "DCC << " << msg;
      #endif
      send(plug, msg.c_str(), msg.length(), 0);
    }
  } else
    return INVALID;
  return 0;
}


// Write(strings, prefix, suffix)
//
// Sends a list of strings to the DCC connection
//
// Returns 0 if no errors ocurred.
//
int DCC::Write(StrList & msg, const std::string & prefix,
  const std::string & suffix)
{
  for (StrList::iterator pos = msg.begin(); pos != msg.end(); ++pos)
  {
    if (0 != Write(prefix + *pos + suffix))
    {
      return INVALID;
    }
  }
  return 0;
}


// PrivSetFD(readset, writeset)
//
// Duh!
//
void DCC::PrivSetFD(fd_set & readset, fd_set & writeset) const
{
  if (listening || connected)
    FD_SET(plug, &readset);
  else
    FD_SET(plug, &writeset);
}


// PrivProcess(readset, writeset)
//
// Processes any incoming data
//
// Returns 0 if no errors ocurred
//
int DCC::PrivProcess(const fd_set & readset, const fd_set & writeset)
{
  std::string text;

  if (connected) {
    if (FD_ISSET(plug, &readset)) {
      if (int i = ReadLine(plug, text) <= 0)
        return INVALID;
      if (text != "") {
        LastActive = time((time_t *)NULL);
        #ifdef DCC_DEBUG
          cout << "DCC >> " << text << endl;
        #endif
        return Parse(text);
      }
    }
    if (!Authed && (GetIdle() > DCC_IDLE_MAX))
      return INVALID;
  } else if (listening) {
    struct sockaddr remotehost;
    int newplug;
    socklen_t size;

    if (FD_ISSET(plug, &readset)) {
      newplug = accept(plug, &remotehost, &size);
      if (newplug > 0) {
	close(plug);
	plug = newplug;
	listening = false;
        connected = true;
        Write("OOMon-" + std::string(OOMON_VERSION) + " by Timothy L. Jensen\n");
        MOTD();
        LastActive = time((time_t *)NULL);
        DCC::Status("*", GetUserHost() + " has connected", this);
        Remote::Status("*", GetUserHost() + " has connected");
      }
    }
    if (GetIdle() > DCC_IDLE_MAX) {
      close(plug);
      return INVALID;
    }
  } else {
    if (FD_ISSET(plug, &writeset)) {
      connected = true;
      Write("OOMon-" + std::string(OOMON_VERSION) + " by Timothy L. Jensen\n");
      MOTD();
      LastActive = time((time_t *)NULL);
      DCC::Status("*", GetUserHost() + " has connected", this);
      Remote::Status("*", GetUserHost() + " has connected");
    }
    if (GetIdle() > DCC_IDLE_MAX) {
      close(plug);
      return INVALID;
    }
  }
  return 0;
}


// Parse(text)
//
// This is where we actually parse the data received. Fun stuff.
//
// Returns 0 in all cases except when the user closes the DCC connection
//
int DCC::Parse(std::string text)
{
  std::string Command, To;
  int AtSign;

  if (text == "")
    return 0;

  if (text[0] == '.') {		// A command?
    // Command
    text.erase((std::string::size_type) 0, 1);
    Command = FirstWord(text);
    AtSign = Command.find('@');  		// Is this a remote command?
    if (AtSign != std::string::npos) {
      if ((Flags & UF_REMOTE) == 0) {  		// Yes! Do I have permission?
        Write("*** You can't use remote commands!\n");
	return 0;
      }
      To = Command.substr(AtSign + 1);
      Command = Command.substr(0, AtSign);
      #ifdef DCC_DEBUG
	cout << "Remote command for " << To << ": " << Command << endl;
      #endif
    } else {
      To = "";
      #ifdef DCC_DEBUG
	cout << "Local command: " << Command << endl;
      #endif
    }
    if (To != "")
      if (Same(To, Config::GetNick())) 	// No need to route to myself!
        To = "";
      else
	if (!Remote::FindBot(To)) {
	  Write("*** Bot not linked: " + To + "\n");
	  return 0;
        }

    // This is a bit messy...
    switch(DCCCommand(UpCase(Command))) {
    case DCC_AUTH:
      if (To == "") {
	std::string Parm1 = FirstWord(text);
	if (Config::Auth(Parm1, UserHost, text, Flags, Handle)) {
	  Authed = true;
	  Write("*** Authorization granted!\n");
	  Status(Handle + " (" + UserHost + ") has been authorized");
	  Log::Write(Handle + " (" + UserHost + ") has been authorized");
	} else {
	  Write("*** Authorization denied!\n");
	  Status(UserHost + " has failed authorization!");
	  Log::Write(UserHost + " has failed authorization!");
	}
      }
      break;
    case DCC_QUIT:
      if (To == "")
        return INVALID;
      break;
    case DCC_CHAT:
      if (To == "")
	Chat(text);
      break;
    case DCC_WHO:
      if (Authed)
      {
        if (To == "")
        {
          StrList Output;
	  Who(Output);
	  Write(Output);
        }
        else
        {
          Remote::Who(Handle, To);
        }
      }
      break;
    case DCC_LINKS:
      if (Authed)
      {
        if (To == "")
        {
          StrList Output;
	  Remote::GetLinks(Output);
	  Write(Output);
        }
        else
        {
	  Remote::DoLinks(Handle, To);
        }
      }
      break;
    case DCC_HELP:
      if (To == "")
        Help(text);
      break;
    case DCC_WATCH:
      if (Authed)
      {
      }
      break;
    case DCC_ECHO:
      {
	std::string Parm = UpCase(FirstWord(text));
	if (Parm == "")
	{
	  Write(std::string("*** ECHO is ") +
	    (echoMyChatter ? "ON" : "OFF") + ".\n");
	}
	else if (Parm == "ON")
	{
	  echoMyChatter = true;
	  Write("*** ECHO is now ON.\n");
	}
	else if (Parm == "OFF")
	{
	  echoMyChatter = false;
	  Write("*** ECHO is now OFF.\n");
	}
	else
	{
	  Write("*** Huh?  Turn echo ON or OFF?\n");
	}
      }
      break;
    case DCC_OP:
      if (Flags & UF_CHANOP) {
	std::string Parm1 = FirstWord(text);
	std::string Parm2 = FirstWord(text);
	if (Parm1 != "")
	  if (Parm2 != "")
	    IRC::Op(Parm1, Parm2);
	  else
	    IRC::Op(Parm1, Nick);
	else
	  Write("*** Please include a channel\n");
      } else
	NoAccess();
      break;
    case DCC_JOIN:
      if (Flags & UF_CHANOP) {
	std::string Parm1 = FirstWord(text);
	std::string Parm2 = FirstWord(text);
	if (Parm1 != "")
	  IRC::Join(Parm1, Parm2);
	else
	  Write("*** Please include a channel\n");
      } else
	NoAccess();
      break;
    case DCC_PART:
      if (Flags & UF_CHANOP) {
	std::string Parm1 = FirstWord(text);
	if (Parm1 != "")
	  IRC::Part(Parm1);
	else
	  Write("*** Please include a channel\n");
      } else
	NoAccess();
      break;
    case DCC_KLINE:
      if (Flags & UF_KLINE) {
	unsigned int Minutes = 0;
	std::string Parm1 = FirstWord(text);
	if ((Minutes = atoi(Parm1.c_str())) != 0)
	  Parm1 = FirstWord(text);
	if (Parm1 != "")
	  if (text != "")
	    if (To == "")
	      Kline(Minutes, Parm1, text);
	    else
	      Remote::Kline(Handle, To, Minutes, Parm1, text);
	  else
	    Write("*** Please include a kline reason\n");
	else
	  Write("*** Please include a nick or user@host mask to kline\n");
      } else
	NoAccess();
      break;
    case DCC_UNKLINE:
      if (Flags & UF_KLINE) {
	std::string Parm1 = FirstWord(text);
	if (Parm1 != "")
	  if (To == "")
	    Unkline(Parm1);
	  else
	    Remote::Unkline(Handle, To, Parm1);
	else
	  Write("*** Please include a user@host mask to unkline\n");
      } else
	NoAccess();
      break;
    case DCC_KCLONE:
      if (Flags & UF_KLINE) {
	unsigned int Minutes = 0;
	std::string Parm1 = FirstWord(text);
	if ((Minutes = atoi(Parm1.c_str())) != 0)
	  Parm1 = FirstWord(text);
	if (Parm1 != "")
	  if (To == "")
	    KClone(Minutes, Parm1);
	  else
	    Remote::KClone(Handle, To, Minutes, Parm1);
	else
	  Write("*** Please include a nick or user@host mask to kline\n");
      } else
	NoAccess();
      break;
    case DCC_KFLOOD:
      if (Flags & UF_KLINE) {
	unsigned int Minutes = 0;
	std::string Parm1 = FirstWord(text);
	if ((Minutes = atoi(Parm1.c_str())) != 0)
	  Parm1 = FirstWord(text);
	if (Parm1 != "")
	  if (To == "")
	    KFlood(Minutes, Parm1);
	  else
	    Remote::KFlood(Handle, To, Minutes, Parm1);
	else
	  Write("*** Please include a nick or user@host mask to kline\n");
      } else
	NoAccess();
      break;
    case DCC_KBOT:
      if (Flags & UF_KLINE) {
	unsigned int Minutes = 0;
	std::string Parm1 = FirstWord(text);
	if ((Minutes = atoi(Parm1.c_str())) != 0)
	  Parm1 = FirstWord(text);
	if (Parm1 != "")
	  if (To == "")
	    KBot(Minutes, Parm1);
	  else
	    Remote::KBot(Handle, To, Minutes, Parm1);
	else
	  Write("*** Please include a nick or user@host mask to kline\n");
      } else
	NoAccess();
      break;
    case DCC_KLINK:
      if (Flags & UF_KLINE) {
	unsigned int Minutes = 0;
	std::string Parm1 = FirstWord(text);
	if ((Minutes = atoi(Parm1.c_str())) != 0)
	  Parm1 = FirstWord(text);
	if (Parm1 != "")
	  if (To == "")
	    KLink(Minutes, Parm1);
	  else
	    Remote::KLink(Handle, To, Minutes, Parm1);
	else
	  Write("*** Please include a nick or user@host mask to kline\n");
      } else
	NoAccess();
      break;
    case DCC_KSPAM:
      if (Flags & UF_KLINE) {
	unsigned int Minutes = 0;
	std::string Parm1 = FirstWord(text);
	if ((Minutes = atoi(Parm1.c_str())) != 0)
	  Parm1 = FirstWord(text);
	if (Parm1 != "")
	  if (To == "")
	    KSpam(Minutes, Parm1);
	  else
	    Remote::KSpam(Handle, To, Minutes, Parm1);
	else
	  Write("*** Please include a nick or user@host mask to kline\n");
      } else
	NoAccess();
      break;
    case DCC_KTRACE:
      if (Flags & UF_KLINE) {
	unsigned int Minutes = 0;
	std::string Parm1 = FirstWord(text);
	if ((Minutes = atoi(Parm1.c_str())) != 0)
	  Parm1 = FirstWord(text);
	if (Parm1 != "")
	  if (To == "")
	    KTrace(Minutes, Parm1);
	  else
	    Remote::KTrace(Handle, To, Minutes, Parm1);
	else
	  Write("*** Please include a nick or user@host mask to kline\n");
      } else
	NoAccess();
      break;
    case DCC_KMOTD:
      if (Flags & UF_KLINE) {
	unsigned int Minutes = 0;
	std::string Parm1 = FirstWord(text);
	if ((Minutes = atoi(Parm1.c_str())) != 0)
	  Parm1 = FirstWord(text);
	if (Parm1 != "")
	  if (To == "")
	    KMotd(Minutes, Parm1);
	  else
	    Remote::KMotd(Handle, To, Minutes, Parm1);
	else
	  Write("*** Please include a nick or user@host mask to kline\n");
      } else
	NoAccess();
      break;
    case DCC_KINFO:
      if (Flags & UF_KLINE) {
	unsigned int Minutes = 0;
	std::string Parm1 = FirstWord(text);
	if ((Minutes = atoi(Parm1.c_str())) != 0)
	  Parm1 = FirstWord(text);
	if (Parm1 != "")
	  if (To == "")
	    KInfo(Minutes, Parm1);
	  else
	    Remote::KInfo(Handle, To, Minutes, Parm1);
	else
	  Write("*** Please include a nick or user@host mask to kline\n");
      } else
	NoAccess();
      break;
    case DCC_KWINGATE:
      if (Flags & UF_KLINE) {
	unsigned int Minutes = 0;
	std::string Parm1 = FirstWord(text);
	if ((Minutes = atoi(Parm1.c_str())) != 0)
	  Parm1 = FirstWord(text);
	if (Parm1 != "")
	  if (To == "")
	    KWingate(Minutes, Parm1);
	  else
	    Remote::KWingate(Handle, To, Minutes, Parm1);
	else
	  Write("*** Please include a nick or user@host mask to kline\n");
      } else
	NoAccess();
      break;
    case DCC_KPROXY:
      if (Flags & UF_KLINE) {
	unsigned int Minutes = 0;
	std::string Parm1 = FirstWord(text);
	if ((Minutes = atoi(Parm1.c_str())) != 0)
	  Parm1 = FirstWord(text);
	if (Parm1 != "")
	  if (To == "")
	    KProxy(Minutes, Parm1);
	  else
	    Remote::KProxy(Handle, To, Minutes, Parm1);
	else
	  Write("*** Please include a nick or user@host mask to kline\n");
      } else
	NoAccess();
      break;
    case DCC_KPERM:
      if (Flags & UF_KLINE) {
	std::string Parm1 = FirstWord(text);
	if (Parm1 != "")
	  if (text != "")
	    if (To == "")
	      KPerm(Parm1, text);
	    else
	      Remote::KPerm(Handle, To, Parm1, text);
	  else
	    Write("*** Please include a kline reason\n");
	else
	  Write("*** Please include a nick or user@host mask to kline\n");
      } else
	NoAccess();
      break;
    case DCC_DLINE:
      if (Flags & UF_DLINE) {
	std::string Parm1 = FirstWord(text);
	if (Parm1 != "")
	  if (text != "")
	    if (To == "")
	      Dline(Parm1, text);
	    else
	      Remote::Dline(Handle, To, Parm1, text);
	  else
	    Write("*** Please include a reason for the dline\n");
	else
	  Write("*** Please include an IP mask to dline\n");
      } else
	NoAccess();
      break;
    case DCC_UNDLINE:
      if (Flags & UF_DLINE) {
	std::string Parm1 = FirstWord(text);
	if (Parm1 != "")
	  if (To == "")
	    Undline(Parm1);
	  else
	    Remote::Undline(Handle, To, Parm1);
	else
	  Write("*** Please include a user@host mask to undline\n");
      } else
	NoAccess();
      break;
    case DCC_GLINE:
      if (Flags & UF_GLINE) {
	if (To == "") {
	  unsigned int Minutes = 0;
	  std::string Parm1 = FirstWord(text);
	  if ((Minutes = atoi(Parm1.c_str())) != 0)
	    Parm1 = FirstWord(text);
	  if (Parm1 != "")
	    if (text != "")
	      Gline(Minutes, Parm1, text);
	    else
	      Write("*** Please include a gline reason\n");
	  else
	    Write("*** Please include a nick or user@host mask to gline\n");
	} else
	  NotRemote();
      } else
	NoAccess();
      break;
    case DCC_UNGLINE:
      if (Flags & UF_GLINE) {
	if (To == "") {
	  std::string Parm1 = FirstWord(text);
	  if (Parm1 != "")
	    Ungline(Parm1);
	  else
	    Write("*** Please include a user@host mask to gline\n");
	} else
	  NotRemote();
      } else
	NoAccess();
      break;
    case DCC_GCLONE:
      if (Flags & UF_GLINE) {
	if (To == "") {
	  unsigned int Minutes = 0;
	  std::string Parm1 = FirstWord(text);
	  if ((Minutes = atoi(Parm1.c_str())) != 0)
	    Parm1 = FirstWord(text);
	  if (Parm1 != "")
	    Gline(Minutes, Parm1, REASON_CLONES);
	  else
	    Write("*** Please include a nick or user@host mask to gline\n");
	} else
	  NotRemote();
      } else
	NoAccess();
      break;
    case DCC_GFLOOD:
      if (Flags & UF_GLINE) {
	if (To == "") {
	  unsigned int Minutes = 0;
	  std::string Parm1 = FirstWord(text);
	  if ((Minutes = atoi(Parm1.c_str())) != 0)
	    Parm1 = FirstWord(text);
	  if (Parm1 != "")
	    Gline(Minutes, Parm1, REASON_FLOODING);
	  else
	    Write("*** Please include a nick or user@host mask to gline\n");
	} else
	  NotRemote();
      } else
	NoAccess();
      break;
    case DCC_GBOT:
      if (Flags & UF_GLINE) {
	if (To == "") {
	  unsigned int Minutes = 0;
	  std::string Parm1 = FirstWord(text);
	  if ((Minutes = atoi(Parm1.c_str())) != 0)
	    Parm1 = FirstWord(text);
	  if (Parm1 != "")
	    Gline(Minutes, Parm1, REASON_BOTS);
	  else
	    Write("*** Please include a nick or user@host mask to gline\n");
	} else
	  NotRemote();
      } else
	NoAccess();
      break;
    case DCC_GLINK:
      if (Flags & UF_GLINE) {
	if (To == "") {
	  unsigned int Minutes = 0;
	  std::string Parm1 = FirstWord(text);
	  if ((Minutes = atoi(Parm1.c_str())) != 0)
	    Parm1 = FirstWord(text);
	  if (Parm1 != "")
	    Gline(Minutes, Parm1, REASON_LINKS);
	  else
	    Write("*** Please include a nick or user@host mask to gline\n");
	} else
	  NotRemote();
      } else
	NoAccess();
      break;
    case DCC_GSPAM:
      if (Flags & UF_GLINE) {
	if (To == "") {
	  unsigned int Minutes = 0;
	  std::string Parm1 = FirstWord(text);
	  if ((Minutes = atoi(Parm1.c_str())) != 0)
	    Parm1 = FirstWord(text);
	  if (Parm1 != "")
	    Gline(Minutes, Parm1, REASON_SPAM);
	  else
	    Write("*** Please include a nick or user@host mask to gline\n");
	} else
	  NotRemote();
      } else
	NoAccess();
      break;
    case DCC_GTRACE:
      if (Flags & UF_GLINE) {
	if (To == "") {
	  unsigned int Minutes = 0;
	  std::string Parm1 = FirstWord(text);
	  if ((Minutes = atoi(Parm1.c_str())) != 0)
	    Parm1 = FirstWord(text);
	  if (Parm1 != "")
	    Gline(Minutes, Parm1, REASON_TRACE);
	  else
	    Write("*** Please include a nick or user@host mask to gline\n");
	} else
	  NotRemote();
      } else
	NoAccess();
      break;
    case DCC_GMOTD:
      if (Flags & UF_GLINE) {
	if (To == "") {
	  unsigned int Minutes = 0;
	  std::string Parm1 = FirstWord(text);
	  if ((Minutes = atoi(Parm1.c_str())) != 0)
	    Parm1 = FirstWord(text);
	  if (Parm1 != "")
	    Gline(Minutes, Parm1, REASON_MOTD);
	  else
	    Write("*** Please include a nick or user@host mask to gline\n");
	} else
	  NotRemote();
      } else
	NoAccess();
      break;
    case DCC_GINFO:
      if (Flags & UF_GLINE) {
	if (To == "") {
	  unsigned int Minutes = 0;
	  std::string Parm1 = FirstWord(text);
	  if ((Minutes = atoi(Parm1.c_str())) != 0)
	    Parm1 = FirstWord(text);
	  if (Parm1 != "")
	    Gline(Minutes, Parm1, REASON_INFO);
	  else
	    Write("*** Please include a nick or user@host mask to gline\n");
	} else
	  NotRemote();
      } else
	NoAccess();
      break;
    case DCC_GWINGATE:
      if (Flags & UF_GLINE) {
	if (To == "") {
	  unsigned int Minutes = 0;
	  std::string Parm1 = FirstWord(text);
	  if ((Minutes = atoi(Parm1.c_str())) != 0)
	    Parm1 = FirstWord(text);
	  if (Parm1 != "")
	    Gline(Minutes, Parm1, REASON_WINGATE);
	  else
	    Write("*** Please include a nick or user@host mask to gline\n");
	} else
	  NotRemote();
      } else
	NoAccess();
      break;
    case DCC_GPROXY:
      if (Flags & UF_GLINE) {
	if (To == "") {
	  unsigned int Minutes = 0;
	  std::string Parm1 = FirstWord(text);
	  if ((Minutes = atoi(Parm1.c_str())) != 0)
	    Parm1 = FirstWord(text);
	  if (Parm1 != "")
	    Gline(Minutes, Parm1, REASON_PROXY);
	  else
	    Write("*** Please include a nick or user@host mask to gline\n");
	} else
	  NotRemote();
      } else
	NoAccess();
      break;
    case DCC_GPERM:
      if (Flags & UF_GLINE) {
	if (To == "") {
	  std::string Parm1 = FirstWord(text);
	  if (Parm1 != "")
	    if (text != "")
	      Gline(0, Parm1, "PERM " + text);
	    else
	      Write("*** Please include a gline reason\n");
	  else
	    Write("*** Please include a nick or user@host mask to gline\n");
	} else
	  NotRemote();
      } else
	NoAccess();
      break;
    case DCC_KILL:
      if (Flags & UF_OPER) {
	if (To == "") {
	  std::string Parm1 = FirstWord(text);
	  if (Parm1 != "") {
	    if (text != "") {
	      Status("KILL " + Parm1 + " ((" + Handle + "@" + Config::GetNick() + ") " + text + ")");
	      IRC::Kill(Handle + "@" + Config::GetNick(), Parm1, text);
	    } else
	      Write("*** Please include a kill reason.\n");
	  } else
	    Write("*** Please include a nick to kill.\n");
	}
      } else
	NoAccess();
      break;
    case DCC_KILLLIST:
      if (Flags & UF_OPER) {
	if (To == "") {
	  int Count;
	  std::string Parm1 = FirstWord(text);
	  if ((Parm1 == "*") || (Parm1 == "*@*")) {
	    Write("Killing all users is not recommended. To do it anyway,\n");
	    Write("use '.killlist ?*@*'.\n");
	    break;
	  }
	  if (Parm1 != "") {
	    if (text != "") {
	      Status("KILLLIST " + Parm1 + " (" + text + ") [" + Handle + "@" + Config::GetNick() + "]");
              Kill_List_Users(Handle + "@" + Config::GetNick(), Parm1, text);
	    }
	    else
	    {
	      Status("KILLLIST " + Parm1 + " [" + Handle + "@" + Config::GetNick() + "]");
              Kill_List_Users(Handle + "@" + Config::GetNick(), Parm1);
	    }
	  } else
	    Write("*** Please include a user@host mask to kill.\n");
	}
      } else
	NoAccess();
      break;
    case DCC_MULTI:
      if (Flags & UF_OPER) {
	if (To == "") {
	  std::string Parm1 = FirstWord(text);
	  int min = atoi(Parm1.c_str());
	  StrList Output;
	  Report_Multi(Output, min);
	  Write(Output);
	}
      } else
	NoAccess();
      break;
    case DCC_HMULTI:
      if (Flags & UF_OPER) {
	if (To == "") {
	  std::string Parm1 = FirstWord(text);
	  int min = atoi(Parm1.c_str());
	  StrList Output;
	  Report_HMulti(Output, min);
	  Write(Output);
	}
      } else
	NoAccess();
      break;
    case DCC_UMULTI:
      if (Flags & UF_OPER) {
	if (To == "") {
	  std::string Parm1 = FirstWord(text);
	  int min = atoi(Parm1.c_str());
	  StrList Output;
	  Report_UMulti(Output, min);
	  Write(Output);
	}
      } else
	NoAccess();
      break;
    case DCC_VMULTI:
      if (Flags & UF_OPER) {
	if (To == "") {
	  std::string Parm1 = FirstWord(text);
	  int min = atoi(Parm1.c_str());
	  StrList Output;
	  Report_VMulti(Output, min);
	  Write(Output);
	}
      } else
	NoAccess();
      break;
    case DCC_NFIND:
      if (Flags & UF_OPER) {
	if (To == "") {
	  std::string Parm1 = FirstWord(text);
	  if (Parm1 != "") {
	    StrList Output;
	    List_Nicks(Output, Parm1);
	    Write(Output);
	  } else
	    Write("*** Please include a nick mask to find.\n");
	}
      } else
	NoAccess();
      break;
    case DCC_TRACE:
      if (Flags & UF_OPER) {
	if (To == "") {
	  IRC::Retrace(Handle + "@" + Config::GetNick());
	}
      } else
	NoAccess();
      break;
    case DCC_CLONES:
      if (Flags & UF_OPER) {
	if (To == "") {
	  StrList Output;
	  Report_Clones(Output);
	  Write(Output);
	}
      } else
	NoAccess();
      break;
    case DCC_LIST:
      if (Flags & UF_OPER) {
	if (To == "") {
	  std::string Parm1 = FirstWord(text);
	  if (Parm1 != "") {
	    StrList Output;
	    List_Users(Output, Parm1);
	    Write(Output);
	  } else
	    Write("*** Please include a user@host mask to list\n");
	}
      } else
	NoAccess();
      break;
    case DCC_IPLIST:
      if (Flags & UF_OPER)
      {
	if (To == "")
	{
	  std::string Parm1 = FirstWord(text);
	  if (Parm1 != "")
	  {
	    StrList Output;
	    List_IPs(Output, Parm1);
	    Write(Output);
	  }
	  else
	  {
	    Write("*** Please include an IP mask to list\n");
	  }
	}
      } else
	NoAccess();
      break;
    case DCC_FINDK:
      if (Flags & UF_OPER) {
	if (To == "") {
	  std::string Parm1 = FirstWord(text);
	  if (Parm1 != "")
	  {
	    StrList Output;
	    IRC::FindK(Output, Parm1);
	    Write(Output);
	  }
	  else
	  {
	    Write("*** Please include a user@host mask to list klines for\n");
	  }
	}
      } else
	NoAccess();
      break;
    case DCC_FINDD:
      if (Flags & UF_OPER) {
	if (To == "") {
	  std::string Parm1 = FirstWord(text);
	  if (Parm1 != "")
	  {
	    StrList Output;
	    IRC::FindD(Output, Parm1);
	    Write(Output);
	  }
	  else
	  {
	    Write("*** Please include an IP mask to list dlines for\n");
	  }
	}
      } else
	NoAccess();
      break;
    case DCC_RELOAD:
      if (Flags & UF_OPER) {
	if (To == "") {
	  std::string Parm1 = UpCase(FirstWord(text));
	  if (Parm1 == "USERS") {
	    IRC::Retrace(Handle + "@" + Config::GetNick());
	  } else if (Parm1 == "KLINES") {
	    IRC::ReloadKlines(Handle + "@" + Config::GetNick());
	  } else if (Parm1 == "DLINES") {
	    IRC::ReloadDlines(Handle + "@" + Config::GetNick());
	  } else if (Parm1 == "CONFIG") {
	    ReloadConfig(Handle + "@" + Config::GetNick());
	  } else if (Parm1 == "HELP") {
	    Help::flush();
	  } else
	    Write("*** Valid options are USERS, KLINES, CONFIG, and HELP\n");
	}
      } else
	NoAccess();
      break;
    case DCC_DOMAINS:
      if (Flags & UF_OPER) {
	if (To == "") {
          std::string Parm1 = FirstWord(text);
	  int min = 5;
	  if (Parm1 != "")
	    min = atoi(Parm1.c_str());
	  if (min >= 1) {
	    StrList Output;
	    Report_Domains(Output, min);
	    Write(Output);
	  } else
	    Write("*** Invalid minimum user size!\n");
	}
      } else
	NoAccess();
      break;
    case DCC_CLASS:
      if (Flags & UF_OPER) {
	if (To == "") {
          std::string Parm1 = FirstWord(text);
	  StrList Output;
	  if (Parm1 != "")
	    CountClasses(Output, Parm1);
	  else
	    CountClasses(Output);
	  Write(Output);
	}
      } else
	NoAccess();
      break;
    case DCC_MOTD:
      if (Authed) {
	if (To == "") {
	  MOTD();
	}
      }
      break;
    case DCC_CONN:
      if (Flags & UF_MASTER) {
	std::string Parm1 = FirstWord(text);
	if (Parm1 != "")
	  Remote::Conn(Handle, To, Parm1);
	else
	  Write("*** Please specify the bot to connect\n");
      } else
	NoAccess();
      break;
    case DCC_DISCONN:
      if (Flags & UF_MASTER) {
	std::string Parm1 = FirstWord(text);
	if (Parm1 != "")
	  Remote::Disconn(Handle, To, Parm1);
	else
	  Write("*** Please specify the bot to disconnect\n");
      } else
	NoAccess();
      break;
    case DCC_DIE:
      if (Flags & UF_MASTER) {
	if (To == "") {
	  Status(Handle + "@" + Config::GetNick() + " says I have to die now...");
	  Write("As you wish...\n");
	  Log::Write("DIE requested by " + Handle + "@" + Config::GetNick());
	  gracefuldie(SIGTERM); 
	} else
	  Remote::Die(Handle, To);
      } else
	NoAccess();
      break;
    case DCC_RAW:
      if (Flags & UF_MASTER) {
	if (To == "") {
	  //Status(Handle + "@" + Config::GetNick() + " says I have to die now...");
	  //Write("As you wish...\n");
	  Log::Write("RAW by " + Handle + "@" + Config::GetNick() + ": " + text);
	  IRC::Write(text + "\n");
	} //else
	  //Remote::Die(Handle, To);
      } else
	NoAccess();
      break;
    case DCC_TRAP:
      if (Flags & UF_MASTER)
      {
	TrapCommand(text);
      }
      else
      {
	NoAccess();
      }
      break;
    case DCC_AUTOPILOT:
      if (Flags & UF_MASTER) {
        if (To == "") {
#ifdef AUTOPILOT
          std::string Parm1 = UpCase(FirstWord(text));
	  if (Parm1 == "") {
	    if (GetAutoPilot())
	      Write("*** Autopilot is ON\n");
	    else
              Write("*** Autopilot is OFF\n");
	  } else
            if (Parm1 == "ON")
	      SetAutoPilot(Handle + "@" + Config::GetNick(), true);
	    else if (Parm1 == "OFF")
	      SetAutoPilot(Handle + "@" + Config::GetNick(), false);
	    else
	      Write("*** Valid settings are ON and OFF\n");
#else
          Write("*** AUTOPILOT has been disabled.\n");
#endif
         }
      } else
	NoAccess();
      break;
#ifdef DEBUG
    case DCC_TEST:
      if (Flags & UF_MASTER)
      {
	std::string ip = FirstWord(text);
	CheckProxy(ip, ip, "TestProxy", "test@test");
      }
      else
      {
	NoAccess();
      }
      break;
#endif
    case DCC_UNKNOWN:
      Write("*** Invalid command: " + Command + "\n");
      break;
    }
  } else {					// Just chatter :P
    // Chat
#ifdef BOOT_BITCHX
    if (text == "Unknown command") {
      Write("*** Your IRC client is broken. Disconnecting...\n");
      return INVALID;
    }
#endif
    Chat(text);
  }
  return 0;
}


// Chat(text)
//
// Handles any text which should be interpretted as chatter.
//
void DCC::Chat(const std::string & text)
{
#ifndef UNAUTHED_MAY_CHAT
  if (Authed) {
#endif
    Chat(Handle + "@" + Config::GetNick(), "*", text,
      echoMyChatter ? NULL : this);
    Remote::Chat(Handle + "@" + Config::GetNick(), "*", text);
#ifndef UNAUTHED_MAY_CHAT
  } else
    Write("*** Only authorized users may chat!\n");
#endif
}


// MOTD()
//
// Writes the Message Of The Day
//
void DCC::MOTD()
{
  StrList Output;
  ::MOTD(Output);
  Write(Output);
}


// NoAccess()
//
// Tells the user that he doesn't have access to use the command he passed
//
void DCC::NoAccess()
{
  Write("*** You don't have access for that command!\n");
}


// NotRemote()
//
// Tells the user that the command he used isn't usable remotely
void DCC::NotRemote()
{
  Write("*** That command can't be used remotely!\n");
}


// Kline(Minutes, Target, Reason)
//
// Writes a kline to the server
//
void DCC::Kline(const int Minutes, const std::string & Target,
	const std::string & Reason)
{
  if (Minutes)
    Status("KLINE " + IntToStr(Minutes) + " " + Target + " (" + Reason +
      " [" + Handle + "@" + Config::GetNick() + "])");
  else
    Status("KLINE " + Target + " (" + Reason + " [" + Handle + "@" +
      Config::GetNick() + "])");
  IRC::Kline(Handle + "@" + Config::GetNick(), Minutes, Target,
    Reason);
}


// KClone(Minutes, Target)
//
// Writes a kline to the server
//
void DCC::KClone(const int Minutes, const std::string & Target)
{
  if (Minutes)
    Status("KCLONE " + IntToStr(Minutes) + " " + Target + " [" +
      Handle + "@" + Config::GetNick() + "]");
  else
    Status("KCLONE " + Target + " [" + Handle + "@" +
      Config::GetNick() + "]");
  IRC::Kline(Handle + "@" + Config::GetNick(), Minutes, Target,
    REASON_CLONES);
}


// KFlood(Minutes, Target)
//
// Writes a kline to the server
//
void DCC::KFlood(const int Minutes, const std::string & Target)
{
  if (Minutes)
    Status("KFLOOD " + IntToStr(Minutes) + " " + Target + " [" +
      Handle + "@" + Config::GetNick() + "]");
  else
    Status("KFLOOD " + Target + " [" + Handle + "@" +
      Config::GetNick() + "]");
  IRC::Kline(Handle + "@" + Config::GetNick(), Minutes, Target,
    REASON_FLOODING);
}


// KBot(Minutes, Target)
//
// Writes a kline to the server
//
void DCC::KBot(const int Minutes, const std::string & Target)
{
  if (Minutes)
    Status("KBOT " + IntToStr(Minutes) + " " + Target + " [" +
      Handle + "@" + Config::GetNick() + "]");
  else
    Status("KBOT " + Target + " [" + Handle + "@" +
      Config::GetNick() + "]");
  IRC::Kline(Handle + "@" + Config::GetNick(), Minutes, Target,
    REASON_BOTS);
}


// KSpam(Minutes, Target)
//
// Writes a kline to the server
//
void DCC::KSpam(const int Minutes, const std::string & Target)
{
  if (Minutes)
    Status("KSPAM " + IntToStr(Minutes) + " " + Target + " [" +
      Handle + "@" + Config::GetNick() + "]");
  else
    Status("KSPAM " + Target + " [" + Handle + "@" +
      Config::GetNick() + "]");
  IRC::Kline(Handle + "@" + Config::GetNick(), Minutes, Target,
    REASON_SPAM);
}


// KLink(Minutes, Target)
//
// Writes a kline to the server
//
void DCC::KLink(const int Minutes, const std::string & Target)
{
  if (Minutes)
    Status("KLINK " + IntToStr(Minutes) + " " + Target + " [" +
      Handle + "@" + Config::GetNick() + "]");
  else
    Status("KLINK " + Target + " [" + Handle + "@" +
      Config::GetNick() + "]");
  IRC::Kline(Handle + "@" + Config::GetNick(), Minutes, Target,
    REASON_LINKS);
}


// KTrace(Minutes, Target)
//
// Writes a kline to the server
//
void DCC::KTrace(const int Minutes, const std::string & Target)
{
  if (Minutes)
    Status("KTRACE " + IntToStr(Minutes) + " " + Target + " [" +
      Handle + "@" + Config::GetNick() + "]");
  else
    Status("KTRACE " + Target + " [" + Handle + "@" +
      Config::GetNick() + "]");
  IRC::Kline(Handle + "@" + Config::GetNick(), Minutes, Target,
    REASON_TRACE);
}


// KMotd(Minutes, Target)
//
// Writes a kline to the server
//
void DCC::KMotd(const int Minutes, const std::string & Target)
{
  if (Minutes)
    Status("KMOTD " + IntToStr(Minutes) + " " + Target + " [" +
      Handle + "@" + Config::GetNick() + "]");
  else
    Status("KMOTD " + Target + " [" + Handle + "@" +
      Config::GetNick() + "]");
  IRC::Kline(Handle + "@" + Config::GetNick(), Minutes, Target,
    REASON_MOTD);
}


// KInfo(Minutes, Target)
//
// Writes a kline to the server
//
void DCC::KInfo(const int Minutes, const std::string & Target)
{
  if (Minutes)
    Status("KINFO " + IntToStr(Minutes) + " " + Target + " [" +
      Handle + "@" + Config::GetNick() + "]");
  else
    Status("KINFO " + Target + " [" + Handle + "@" +
      Config::GetNick() + "]");
  IRC::Kline(Handle + "@" + Config::GetNick(), Minutes, Target,
    REASON_INFO);
}


// KWingate(Minutes, Target)
//
// Writes a kline to the server
//
void DCC::KWingate(const int Minutes, const std::string & Target)
{
  if (Minutes)
    Status("KWINGATE " + IntToStr(Minutes) + " " + Target + " [" +
      Handle + "@" + Config::GetNick() + "]");
  else
    Status("KWINGATE " + Target + " [" + Handle + "@" +
      Config::GetNick() + "]");
  IRC::Kline(Handle + "@" + Config::GetNick(), Minutes, Target,
    REASON_WINGATE);
}


// KProxy(Minutes, Target)
//
// Writes a kline to the server
//
void DCC::KProxy(const int Minutes, const std::string & Target)
{
  if (Minutes)
    Status("KPROXY " + IntToStr(Minutes) + " " + Target + " [" +
      Handle + "@" + Config::GetNick() + "]");
  else
    Status("KPROXY " + Target + " [" + Handle + "@" +
      Config::GetNick() + "]");
  IRC::Kline(Handle + "@" + Config::GetNick(), Minutes, Target,
    REASON_PROXY);
}


// KPerm(Target, Reason)
//
// Writes a kline to the server
//
void DCC::KPerm(const std::string & Target, const std::string & Reason)
{
  Status("KPERM " + Target + " (" + Reason + " [" + Handle +
    "@" + Config::GetNick() + "])");
  IRC::Kline(Handle + "@" + Config::GetNick(), 0, Target,
    "PERM " + Reason);
}


// Unkline(Target)
//
// Just the opposite of Kline
//
void DCC::Unkline(const std::string & Target)
{
  Status("UNKLINE " + Target + " [" + Handle + "@" + Config::GetNick() + "]");
  IRC::Unkline(Handle + "@" + Config::GetNick(), Target);
}


// Dline(Target)
//
// Writes a dline to the server
//
void DCC::Dline(const std::string & Target, const std::string & Reason)
{
  Status("DLINE " + Target + " (" + Reason + " [" + Handle + "@" +
    Config::GetNick() + "])");
  IRC::Dline(Handle + "@" + Config::GetNick(), Target, Reason);
}


// Undline(Target)
//
// Just the opposite of Dline
//
void DCC::Undline(const std::string & Target)
{
  Status("UNDLINE " + Target + " [" + Handle + "@" + Config::GetNick() + "]");
  IRC::Undline(Handle + "@" + Config::GetNick(), Target);
}


// Gline(Minutes, Target, Reason)
//
// Writes a kline too all linked servers
//
void DCC::Gline(const int Minutes, const std::string & Target,
	const std::string & Reason)
{
  if (Minutes)
    Status("GLINE " + IntToStr(Minutes) + " " + Target + " (" +
      Reason + " [" + Handle + "@" + Config::GetNick() + "])");
  else
    Status("GLINE " + Target + " (" + Reason + " [" + Handle + "@" +
      Config::GetNick() + "])");
  IRC::Kline(Handle + "@" + Config::GetNick(), Minutes, Target, Reason);
  Remote::Gline(Handle, Minutes, Target, Reason);
}


// Ungline(Target)
//
// Opposite of Gline (duh)
//
void DCC::Ungline(const std::string & Target)
{
  Status("UNGLINE " + Target + " [" + Handle + "@" + Config::GetNick() + "]");
  IRC::Unkline(Handle + "@" + Config::GetNick(), Target);
  Remote::Ungline(Handle, Target);
}


// Help(Topic)
//
// Gets help
//
void DCC::Help(std::string Topic)
{
  StrList text = Help::getHelp(Topic);

  if (text.size() <= 0)
  {
    Write("Help can be read at http://toast.blackened.com/oomon/\n");
  }
  else
  {
    Write(text);
  }
}


// GetIdle()
//
// Returns the amount of time the user has been idle
//
int DCC::GetIdle() const
{
  time_t CurrentTime;

  CurrentTime = time((time_t *)NULL);
  return (CurrentTime - LastActive);
}


// GetFlags()
//
// Returns a fixed length string containing the user's flags
//
std::string DCC::GetFlags() const
{
  if (Authed) {
    std::string temp = "";
    if (Flags & UF_CHANOP) temp += "C"; else temp += " ";
    if (Flags & UF_DLINE) temp += "D"; else temp += " ";
    if (Flags & UF_GLINE) temp += "G"; else temp += " ";
    if (Flags & UF_KLINE) temp += "K"; else temp += " ";
    if (Flags & UF_MASTER) temp += "M"; else temp += " ";
    if (Flags & UF_OPER) temp += "O"; else temp += " ";
    if (Flags & UF_REMOTE) temp += "R"; else temp += " ";
    if (Flags & UF_WALLOPS) temp += "W"; else temp += " ";
    return temp;
  } else
    return "-       ";
}


// IsOper()
//
// Returns true if user is +O
//
bool DCC::IsOper() const
{
  return ((Flags & UF_OPER) > 0);
}


// Status()
//
// Writes a status message to all connections except this one
//
void DCC::Status(const std::string & Text)
{
  Status("*", Text, this);
  Remote::Status("*", Text);
}

bool DCC::IsConnected() const
{
  return connected;
}


void
DCC::TrapCommand(std::string text)
{
  std::string flag = UpCase(FirstWord(text));
  std::string mask = FirstWord(text);
  std::string timeoutStr;
  long timeout = 0;

  if (atol(mask.c_str()) != 0)
  {
    timeoutStr = mask;
    timeout = atol(timeoutStr.c_str());
    mask = FirstWord(text);
  }

  std::string reason = text;

  if (flag == "" || flag == "LIST")
  {
    StrList output;
    TrapList::list(output);
    Write(output);
  }
  else if (flag == "REMOVE")
  {
    if (TrapList::remove(mask))
    {
      ::Status(Handle + "@" + Config::GetNick() + " removed trap: " + mask,
	SEND_OPERS_ONLY);
      Log::Write(Handle + "@" + Config::GetNick() + " removed trap: " + mask);
    }
    else
    {
      Write(std::string("*** No trap exists for: ") + mask + "\n",
	SEND_OPERS_ONLY);
    }
  }
  else if (flag == "ECHO")
  {
    TrapList::addEcho(mask);
    ::Status(Handle + "@" + Config::GetNick() + " added ECHO trap: " + mask,
      SEND_OPERS_ONLY);
    Log::Write(Handle + "@" + Config::GetNick() + " added ECHO trap: " + mask);
  }
  else if (flag == "KILL")
  {
    TrapList::addKill(mask, reason);
    ::Status(Handle + "@" + Config::GetNick() + " added KILL trap: " + mask +
      " (" + reason + ")", SEND_OPERS_ONLY);
    Log::Write(Handle + "@" + Config::GetNick() + " added KILL trap: " + mask +
      " (" + reason + ")");
  }
  else if (flag == "KLINE")
  {
    TrapList::addKline(false, mask, reason, timeout);
    if (timeout > 0)
    {
      ::Status(Handle + "@" + Config::GetNick() + " added KLINE " +
	timeoutStr + " trap: " + mask + " (" + reason + ")", SEND_OPERS_ONLY);
      Log::Write(Handle + "@" + Config::GetNick() + " added KLINE " +
	timeoutStr + " trap: " + mask + " (" + reason + ")");
    }
    else
    {
      ::Status(Handle + "@" + Config::GetNick() + " added KLINE trap: " +
	mask + " (" + reason + ")", SEND_OPERS_ONLY);
      Log::Write(Handle + "@" + Config::GetNick() + " added KLINE trap: " +
	mask + " (" + reason + ")");
    }
  }
  else if (flag == "KLINEHOST")
  {
    TrapList::addKline(true, mask, reason, timeout);
    if (timeout > 0)
    {
      ::Status(Handle + "@" + Config::GetNick() + " added KLINEHOST " +
	timeoutStr + " trap: " + mask + " (" + reason + ")", SEND_OPERS_ONLY);
      Log::Write(Handle + "@" + Config::GetNick() + " added KLINEHOST " +
	timeoutStr + " trap: " + mask + " (" + reason + ")");
    }
    else
    {
      ::Status(Handle + "@" + Config::GetNick() + " added KLINEHOST trap: " +
	mask + " (" + reason + ")", SEND_OPERS_ONLY);
      Log::Write(Handle + "@" + Config::GetNick() + " added KLINEHOST trap: " +
	mask + " (" + reason + ")");
    }
  }
  else if (flag == "DLINE")
  {
    TrapList::addDline(false, mask, reason);
    ::Status(Handle + "@" + Config::GetNick() + " added DLINE trap: " + mask +
      " (" + reason + ")", SEND_OPERS_ONLY);
    Log::Write(Handle + "@" + Config::GetNick() + " added DLINE trap: " + mask +
      " (" + reason + ")");
  }
  else if (flag == "DLINENET")
  {
    TrapList::addDline(true, mask, reason);
    ::Status(Handle + "@" + Config::GetNick() + " added DLINENET trap: " +
      mask + " (" + reason + ")", SEND_OPERS_ONLY);
    Log::Write(Handle + "@" + Config::GetNick() + " added DLINENET trap: " +
      mask + " (" + reason + ")");
  }
  else
  {
    Write("*** Invalid TRAP action: " + flag + "\n");
  }
}

