// $Id: engine.cc,v 1.67.4.1 2003/01/23 20:47:42 toast Exp $

// This is the beefy part of OOMon.  Most of this was ripped and/or
// ported from Dianora's tcm-043 for the sake of my own sanity (and
// grades).
//
// Eventually I'll go through and try to rewrite some of this. Most
// of the slowdowns are caused by std::string to (char *) conversions
// and the like. I'll try to cut down on those. Otherwise this
// works pretty smoothly.

// C headers
#include <sys/types.h>    
#include <sys/socket.h>    
#include <netinet/in.h>    
#include <arpa/inet.h>
#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <signal.h>
#include <netdb.h>

// C++ headers
#include <string>
#include <map>

#include "strtype"
#include "oomon.h"
#include "log.h"
#include "main.h"
#include "irc.h"
#include "util.h"
#include "config.h"
#include "engine.h"
#include "proxy.h"
#include "trap.h"


#define HASHTABLESIZE 3001

#define CLONECONNECTCOUNT 3
#define CLONECONNECTFREQ  30

#define CLONEDETECTINC 15
#define MAXFROMHOST    50

typedef struct userentry
{
  char nick[MAX_NICK];
  char user[11];
  char host[MAX_HOST];
  char domain[MAX_DOMAIN];
  unsigned long ip;
  char link_count;
  char isoper;
  char userclass[MAX_CLASS];
  time_t connecttime;
  time_t reporttime;
} USERENTRY;

typedef struct hashrec
{
  USERENTRY *info;
  struct hashrec *collision;
} HASHREC;

#ifdef AUTOPILOT
 bool autopilot = AUTOPILOT_DEFAULT;
#endif

HASHREC *hosttable[HASHTABLESIZE];
HASHREC *domaintable[HASHTABLESIZE];
HASHREC *usertable[HASHTABLESIZE];
HASHREC *iptable[HASHTABLESIZE];

typedef struct sortarray
{
  struct userentry *domainrec;
  int count;
} SORTARRAY;

struct failrec
{
  char user[MAX_NICK+1];
  char host[MAX_HOST];
  int botcount;
  int failcount;
  struct failrec *next;
};    

struct failrec *failures = NULL;

#define NICK_CHANGE_TABLE_SIZE 100  

typedef struct nick_change_entry
{ 
  char *user_host;
  char last_nick[MAX_NICK];
  int  nick_change_count;
  time_t first_nick_change;
  time_t last_nick_change;
  bool noticed;  
} NICK_CHANGE_ENTRY;

NICK_CHANGE_ENTRY *nick_changes[NICK_CHANGE_TABLE_SIZE];

#define LINK_LOOK_TABLE_SIZE 10

typedef struct link_look_entry
{
  char *user_host;
  int  link_look_count;
  time_t last_link_look;
} LINK_LOOK_ENTRY;

LINK_LOOK_ENTRY link_look[LINK_LOOK_TABLE_SIZE];

#define TRACE_TABLE_SIZE 75

typedef struct trace_entry
{
  char *user_host;
  int  trace_count;
  time_t last_trace;
} TRACE_ENTRY;

TRACE_ENTRY trace[TRACE_TABLE_SIZE];

#define MOTD_TABLE_SIZE 50

typedef struct motd_entry
{
  char *user_host;
  int  motd_count;
  time_t last_motd;
} MOTD_ENTRY;

MOTD_ENTRY motd[MOTD_TABLE_SIZE];

#define INFO_TABLE_SIZE 50

typedef struct info_entry
{
  char *user_host;
  int  info_count;
  time_t last_info;
} INFO_ENTRY;

INFO_ENTRY info[INFO_TABLE_SIZE];

#define IP_TABLE_SIZE 300
typedef struct ip_entry
{
  unsigned long	net;
  int		ip_count, total_count;
  time_t	last_ip, first_ip;
} IP_ENTRY;

IP_ENTRY ips[IP_TABLE_SIZE];

unsigned long vhost_netmask = inet_addr(VHOST_NETMASK);

bool DoingTrace = false;


static void Add_IP(std::string ip);
static int Hash_Func(const char *);



static void
Kill_Spoof(const std::string & nick, std::string user_host, const std::string & IP)
{
  // Don't kill our happy fun oper spoofers :)
  if (Config::IsSpoofer(IP))
    return;

  if (IP != "") {
    std::string Hostname = user_host.substr(user_host.find('@') + 1);
    IRC::Kill("Auto-Kill", nick, "DNS spoof (" + IP + " != " + Hostname + ")");
  } else
    IRC::Kill("Auto-Kill", nick, "DNS spoof");
  if (IP != "") {
    ::Status("DNS spoofer: " + nick + " (" + user_host + ") != " + IP,
      SEND_OPERS_ONLY);
    Log::Write("DNS spoofer: " + nick + " (" + user_host + ") != " + IP);
  } else {
    ::Status("DNS spoofer: " + nick + " (" + user_host + ")",
      SEND_OPERS_ONLY);
    Log::Write("DNS spoofer: " + nick + " (" + user_host + ")");
  }
}

static std::string
GetReason(const int ReasonType)
{
  switch (ReasonType) {
  case MK_CLONES:
    return REASON_CLONES;
  case MK_FLOODING:
    return REASON_FLOODING;
  case MK_BOTS:
    return REASON_BOTS;
  case MK_SPAM:
    return REASON_SPAM;
  case MK_LINKS:
    return REASON_LINKS;
  case MK_TRACE:
    return REASON_TRACE;
  case MK_MOTD:
    return REASON_MOTD;
  case MK_INFO:
    return REASON_INFO;
  case MK_WINGATE:
    return REASON_WINGATE;
  case MK_PROXY:
    return REASON_PROXY;
  default:
    return "Undefined Reason";
  }
}

std::string Make_Kline(const std::string & Mask, const int ReasonType, const int dur = 0)
{
  std::string Dur;
  if (dur)
    Dur = IntToStr(dur) + " ";
  else
    Dur = "";
  switch (ReasonType) {
  case MK_CLONES:
    return ".kclone " + Dur + Mask;
  case MK_FLOODING:
    return ".kflood " + Dur + Mask;
  case MK_BOTS:
    return ".kbot " + Dur + Mask;
  case MK_SPAM:
    return ".kspam " + Dur + Mask;
  case MK_LINKS:
    return ".klink " + Dur + Mask;
  case MK_TRACE:
    return ".ktrace " + Dur + Mask;
  case MK_MOTD:
    return ".kmotd " + Dur + Mask;
  case MK_INFO:
    return ".kinfo " + Dur + Mask;
  case MK_WINGATE:
    return ".kwingate " + Dur + Mask;
  case MK_PROXY:
    return ".kproxy " + Dur + Mask;
  }
}

void Suggest_Kline(const bool kline, std::string User, char *host,
	const bool different, const bool identd, const int Reason)
{
  std::string Notice;
  std::string UserAtHost;
  std::string Host = host;
  std::string Suggested_Host;
  char *work_host;      /* copy of the host name */
  char *suggested_host; /* what will be suggested as host to kline */
  char *p;
  char *q;
  int number_of_dots=0; /* how many dots in the name ? */
  bool ip_number = true; 
  
  UserAtHost = User + "@" + Host;

  if(strchr(host,'*') || strchr(host,'?'))
  {
    ::Status("Bogus DNS spoofed host " + UserAtHost, SEND_OPERS_ONLY);
    return;
  }

  work_host = strdup(host);
  if (work_host == (char *)NULL) {
    ::Status("Ran out of memory in Suggest_Kline", SEND_OPERS_ONLY);
    Log::Write("Ran out of memory in Suggest_Kline");
    gracefuldie(SIGTERM);
  }

  p = q = work_host;

  while(*p)
    { 
      if(*p == '.')
        number_of_dots++;
      else if( !isdigit(*p) )
        ip_number = false;
      p++;
    } 
    
  if(number_of_dots != 3)
    ip_number = false;
  
  if(ip_number)
    { 
      while(*p != '.')
        {
          p--;
          if( p == q )  /* JUST in case */
            break;
        } 
      *p++ = '.';       /* should already be a '.' right here, but... */
      *p++ = '*';
      *p = '\0';
    } 
  else  
    { 
/*      

*/  
      if(number_of_dots > 1)    /* Is it "some.host.dom" or just
                                   "somehost.dom" ? if only one dot
                                   it's the second form */
        {
          while(*q != '.')
            {
              q++;
              if( *q == '\0' ) /* JUST in case */
                break;
            } 

          p = q;

          while(*p)
            p++;
          while(*p != '.')
            p--;
          p++;
/*      
  I am now at the end of the hostname. the last little bit is the
  top level domain. if its only two letters, then its a country
  domain, and I have to rescan  
*/                       
          if(strlen(p) != 3)    /* *sigh* try again */
            {
              q = work_host;
              if(number_of_dots > 2)
                { 
                  while(*q != '.')
                    { 
                      q++;
                      if( *q == '\0' ) /* JUST in case */
                        break;
                    }
                  q--;
                  *q = '*';
                }
            }
          else
            {
              q--;
              *q = '*';
            } 
        }                
    }     
            
  suggested_host = q;
  Suggested_Host = suggested_host;
                
  if(different)   
    {               
      if(identd)      
        {
                        
  /* added Auto k-line 9/6/96 Phisher dkemp@frontiernet.net */
/* After adding Auto k-line, I thought it might be nice to compare
   first to a hosttable so we aren't klining our own users etc. - Phisher
                
   Wouldn't hurt to make sure they aren't an oper too
   - Dianora  
*/      
         if ( (!Config::IsOKHost(UserAtHost)) && (!Config::IsOper(UserAtHost))) {
#ifdef AUTO_KLINE_HOST
           if (autopilot && kline) {
	     Notice = "Adding temporary Auto Kline for *@" + Host + " :" +
	       GetReason(Reason);
	     IRC::Kline("Auto-Kline", 30, "*@" + Host, GetReason(Reason));
	   } else
#endif
           {
             Notice = Make_Kline("*@" + Host, Reason, 30);
	   }
         } else {
#ifndef DONT_KLINE_OUR_E
              Notice = Make_Kline("*@" + Host, Reason, 30);
#endif
         }
      } else
        { 
          if ( (!Config::IsOKHost(UserAtHost)) && (!Config::IsOper(UserAtHost)) )
            {
#ifdef AUTO_KLINE_NOIDENT
              if(autopilot && kline)
                {
		  Notice = "Adding Auto Kline for ~*@" + Suggested_Host +
		    " :" + GetReason(Reason);
		  IRC::Kline("Auto-Kline", 0, "~*@" + Suggested_Host, GetReason(Reason));
                }
              else
#endif              
                {
		  Notice = Make_Kline("~*@" + Suggested_Host, Reason);
                }
            }
          else
            {   
#ifndef DONT_KLINE_OUR_E
              Notice = Make_Kline("~*@" + Suggested_Host, Reason);
#endif            
            }
        }                
    }     
  else                          
    {
      if (User[0] == '~')
	User.erase((std::string::size_type) 0, 1);
      while (User.length() > 9)
        User.erase((std::string::size_type) 0, 1);
            
      if ( (!Config::IsOKHost(UserAtHost)) && (!Config::IsOper(UserAtHost)) ) {
#ifdef AUTO_KLINE_USERHOST
        if (autopilot && kline) {
	  Notice = "Adding Auto Kline for *" + User + "@" + Suggested_Host +
	    " :" + GetReason(Reason);
	  IRC::Kline("Auto-Kline", 0, "*" + User + "@" +
	    Suggested_Host, GetReason(Reason));
	} else
#endif
        {
          Notice = Make_Kline("*" + User + "@" + Suggested_Host, Reason);
	}
      } else {
#ifndef DONT_KLINE_OUR_E
          Notice = Make_Kline("*" + User + "@" + Suggested_Host, Reason);
#endif
      }
    }      
                        
  if(Notice != "")
    ::Status(Notice, SEND_KLINERS_ONLY);

  (void)free(work_host);
}

void Init_Nick_Change_Table()
{
  for (int i = 0; i < NICK_CHANGE_TABLE_SIZE; i++)
    nick_changes[i] = (NICK_CHANGE_ENTRY *) NULL;
}

static void
Add_To_Nick_Change_Table(char *user_host, char *last_nick)
{
  char notice[MAX_BUFF];
  char *user;
  char *host;
#ifdef AUTO_KILL_NICK_FLOODING
  char hold[MAX_BUFF];
#endif      
  int i;      
  bool found_entry = false;
  int found_empty_entry=INVALID;
  NICK_CHANGE_ENTRY *ncp;
  time_t current_time;
  struct tm *tmrec; 
   
  current_time = time((time_t *)NULL);
   
  for(i = 0; i < NICK_CHANGE_TABLE_SIZE; i++)
    {         
      if(ncp = nick_changes[i])
        {
          time_t time_difference;
          int time_ticks;
  
          time_difference = current_time - ncp->last_nick_change;
  
          /* is it stale ? */ 
          if( time_difference >= NICK_CHANGE_T2_TIME )
            {
              (void)free(ncp->user_host);
              (void)free(nick_changes[i]);
              nick_changes[i] = (NICK_CHANGE_ENTRY *)NULL;
            }
          else
            {
              /* how many 10 second intervals do we have? */
              time_ticks = time_difference / NICK_CHANGE_T1_TIME;
   
              /* is it stale? */
              if(time_ticks >= ncp->nick_change_count)
                {
                  (void)free(ncp->user_host);
                  (void)free(nick_changes[i]);
                  nick_changes[i] = (NICK_CHANGE_ENTRY *)NULL;
                }
              else
                {
                  /* just decrement 10 second units of nick changes */
                  ncp->nick_change_count -= time_ticks;
                  if(strcasecmp(ncp->user_host,user_host) == 0)
                    {
                      ncp->last_nick_change = current_time;
                      strncpy(ncp->last_nick, last_nick, MAX_NICK);
		      ncp->last_nick[MAX_NICK - 1] = 0;
                      ncp->nick_change_count++;
                      found_entry = true;
                    }
              
                  /* now, check for a nick flooder */
   
                  if((ncp->nick_change_count >= NICK_CHANGE_MAX_COUNT)
                     && !ncp->noticed)
                    {
		      (void)sprintf(notice,
                      "nick flood %s (%s) %d in %d seconds",
                                    ncp->user_host,
                                    ncp->last_nick, 
                                    ncp->nick_change_count,
                                    ncp->last_nick_change -
				      ncp->first_nick_change);
                      ::Status(notice, SEND_OPERS_ONLY);
		      Log::Write(notice);
/*                  
   A little convoluted here *sigh*
   but if I want to auto kill nick flooders, I compile
   in a KILL message here, then I still want to suggest a kline to use
                  
  I trust opers will not be nick flooding, but just in case
  I won't AUTO kill an oper..
                      
  - Dianora           
*/                
                      
#ifdef AUTO_KILL_NICK_FLOODING
                      if(autopilot) 
                        {           
                          if ( (!Config::IsOKHost(std::string(user_host))) && (!Config::IsOper(std::string(user_host))))
                            {       
			      IRC::Kill("Auto-Kill", last_nick, "nick flooding");
                            }       
                        }
#endif              
/* 
   get ready to suggest a kline
*/ 
                      user = strtok(user_host,"@");
                      if(user == (char *)NULL)return; /* OOOPSIES! */
                      host = strtok((char *)NULL,"");
                      if(host == (char *)NULL)return; /* OOPSIES! */
  
/*                
   If auto killing nick flooders, suggest the kline but
   tell "suggest_kline()" NOT to kline the nick
*/               
                        
#ifdef AUTO_KILL_NICK_FLOODING 
                      if(*user_host == '~')
                        Suggest_Kline(false,user,host,false,false,MK_FLOODING);
                      else            
                        Suggest_Kline(false,user,host,false,true,MK_FLOODING);
#else                       
/*            
   If NOT auto killing nick flooders, suggest the kline but
   tell "suggest_kline()" TO kline the nick IFF autopilot is ON
   AND AUTO_KLINE is on 
*/ 
                      if(*user_host == '~')
                        Suggest_Kline(true,user,host,false,false,MK_FLOODING);
                      else 
                        Suggest_Kline(true,user,host,false,true,MK_FLOODING);
  
#endif
                      ncp->noticed = true;
                    }               
                }                   
            }                       
        }                           
      else                  
        {     
          if( found_empty_entry == INVALID )
            found_empty_entry = i;  
        }               
    }
                                    
  if (found_entry)       
    return;           
                      
  ncp = (NICK_CHANGE_ENTRY *)malloc(sizeof(NICK_CHANGE_ENTRY));

  if( ncp == (NICK_CHANGE_ENTRY *)NULL)
    {                   
      cerr << "Ran out of memory in add_to_nick_change_table" << endl;
      ::Status("Ran out of memory in add_to_nick_change_table",
	SEND_OPERS_ONLY);
      Log::Write("Ran out of memory in add_to_nick_change_table");
      gracefuldie(SIGTERM);                
    }       
        
  if( (ncp->user_host = strdup(user_host)) == (char *)NULL)
    {   
      ::Status("Ran out of memory in add_to_nick_change_table",
	SEND_OPERS_ONLY);
      Log::Write("Ran out of memory in add_to_nick_change_table");
      gracefuldie(SIGTERM);    
    }
                                    
/* If the table is full, don't worry about this nick change for now
   if this nick change is part of a flood, it will show up
   soon enough anyway... -db
*/

  if(found_empty_entry != INVALID)
    {                   
      ncp->first_nick_change = current_time; 
      ncp->last_nick_change = current_time;
      ncp->nick_change_count = 1;
      ncp->noticed = false;
      nick_changes[found_empty_entry] = ncp;
    }       
}       

void Init_Link_Look_Table()
{
  for (int i = 0; i < LINK_LOOK_TABLE_SIZE; i++)
    link_look[i].user_host = (char *) NULL;
}

void FreeHash()
{
  int i;
  HASHREC *hp;
  HASHREC *next_hp;
  HASHREC *dp;
  HASHREC *next_dp;
  HASHREC *up;
  HASHREC *next_up;

  for (i=0;i<HASHTABLESIZE;++i)
    {
      hp = hosttable[i];
      while(hp)
        {
          next_hp = hp->collision;

          if(hp->info->link_count > 0)
            hp->info->link_count--;

          if(hp->info->link_count == 0)
            {
              (void)free(hp->info);
            }
  
          (void)free(hp);
          hp = next_hp;
        } 
      hosttable[i] = (HASHREC *)NULL;
  
      dp = domaintable[i];
      while(dp)
        {
          next_dp = dp->collision;

          if(dp->info->link_count > 0)
            dp->info->link_count--; 

          if(dp->info->link_count == 0)
            {
              (void)free(dp->info);
            }
  
          (void)free(dp);
          dp = next_dp;
        } 
      domaintable[i] = (HASHREC *)NULL;
  
      up = usertable[i];
      while(up)
        {
          next_up = up->collision;

          if(up->info->link_count > 0)
            up->info->link_count--; 

          if(up->info->link_count == 0)
            {
              (void)free(up->info);
            }
  
          (void)free(up);
          up = next_up;
        } 
      usertable[i] = (HASHREC *)NULL;
  
      up = iptable[i];
      while(up)
        {
          next_up = up->collision;

          if(up->info->link_count > 0)
            up->info->link_count--; 

          if(up->info->link_count == 0)
            {
              (void)free(up->info);
            }
  
          (void)free(up);
          up = next_up;
        } 
      iptable[i] = (HASHREC *)NULL;
    }
}

void InitHash()
{
  for (int i = 0; i < HASHTABLESIZE; i++)
    hosttable[i] = domaintable[i] = usertable[i] = iptable[i] = (HASHREC *)NULL;
  DoingTrace = true;
  IRC::Trace();
}

static void
CheckHostClones(char *host)
{
  HASHREC *find;
  int clonecount = 0; 
  int reportedclones = 0;
  char *last_user="";
  bool last_identd,current_identd;
  bool different;
  time_t now, lastreport, oldest;
  char notice1[MAX_BUFF];
  char notice[MAX_BUFF];
  struct tm *tmrec;
  int index;
  
  oldest = now = time(NULL);
  lastreport = 0;
  index = Hash_Func(host);
  find = hosttable[index];
  
  while (find)
    {
      if (!strcmp(find->info->host,host) &&
          (now - find->info->connecttime < CLONECONNECTFREQ + 1))
        if (find->info->reporttime > 0)
          {
            ++reportedclones;
            if (lastreport < find->info->reporttime)
              lastreport = find->info->reporttime;
          }
        else
          {
            ++clonecount;
            if (find->info->connecttime < oldest)
              oldest = find->info->connecttime;
          }
      find = find->collision;
    }
  
  if ((reportedclones == 0 && clonecount < CLONECONNECTCOUNT) ||
      now - lastreport < 10)
    return;
    
  //initlog();
 
  if (reportedclones)
    {
      (void)sprintf(notice,"%d more possible clones (%d total) from %s:",
            clonecount, clonecount+reportedclones, host);
    } 
  else
    {
      (void)sprintf(notice,
              "Possible clones from %s detected: %d connects in %d seconds",
              host, clonecount, now - oldest);
    } 

  ::Status(notice, SEND_OPERS_ONLY);
  Log::Write(notice);
            
  clonecount = 0;
  find = hosttable[index];
    
  while (find)
    {
    if (!strcmp(find->info->host,host) &&
        (now - find->info->connecttime < CLONECONNECTFREQ + 1) &&
        find->info->reporttime == 0)
      {
        ++clonecount;
        tmrec = localtime(&find->info->connecttime);
          
        if(clonecount == 1)  
          { 
            (void)sprintf(notice1,"  %s is %s@%s (%2.2d:%2.2d:%2.2d)",
                    find->info->nick, find->info->user, find->info->host,
                    tmrec->tm_hour, tmrec->tm_min, tmrec->tm_sec);
          }
        else
          {
            (void)sprintf(notice,"  %s is %s@%s (%2.2d:%2.2d:%2.2d)",
                    find->info->nick, find->info->user, find->info->host,
                    tmrec->tm_hour, tmrec->tm_min, tmrec->tm_sec);
          }
  
        last_identd = current_identd = true;
        different = false;
        
        if(clonecount == 1)
          last_user = find->info->user;
        else if(clonecount == 2)
          {
            char *current_user;
        
            if( *last_user == '~' )
              {
                last_user++;
                last_identd = false;
              }
        
            current_user = find->info->user;
            if( *current_user == '~' ) 
              {     
                current_user++;
                current_identd = false;
              }
        
            if(strcmp(last_user,current_user) != 0)
              different = true;
        
            Suggest_Kline(true,
                          find->info->user,
                          find->info->host,
                          different,
                          last_identd|current_identd,
                          MK_CLONES);
              
          }     
      
        find->info->reporttime = now;
        if(clonecount == 1)
          ; 
        else if(clonecount == 2) 
          {   
            ::Status(notice1, SEND_OPERS_ONLY);
            ::Status(notice, SEND_OPERS_ONLY);
	    Log::Write(notice1);
	    Log::Write(notice);
          }
        else if (clonecount < 5)
          {
            ::Status(notice, SEND_OPERS_ONLY);
	    Log::Write(notice);
          }
        else if (clonecount == 5)
          {
            ::Status(notice, SEND_OPERS_ONLY);
	    Log::Write(notice);
          }
      }
    find = find->collision;
    }
}


// ChopUH(NUH)
//
// Takes the user@host out of a nick[user@host]
// Sorry, I'm too lazy to code the other formats :P
//
static std::string
ChopUH(std::string NUH)
{
  if (NUH != "") {
    std::string::size_type pos = NUH.find('[');
    if (pos != std::string::npos) {
      if (CountChars(NUH, '[') > 1) {
	// Moron has '[' characters in his nick or username
	// Look for '~' to mark the start of an non-ident'ed
	// username...
	std::string::size_type tilde = NUH.find("[~");
	if (tilde != std::string::npos) {
	  NUH = NUH.substr(tilde + 1);
	} else {
	  // Blah. Loser is running identd with a screwy
	  // username or nick.
	  // Find the last '[' in the string and assume it is
	  // the divider, unless this produces an illegal
	  // length nick :P
	  int nicklen = 0;
	  while ((nicklen < 9) && ((pos = NUH.find('[')) != std::string::npos))
	  {
	    NUH = NUH.substr(1);
	    nicklen++;
	  }
	}
      } else {
        NUH = NUH.substr(pos + 1);
      }
      // Remove the trailing ']'
      NUH.erase(NUH.length() - 1, 1);
      return NUH;
    }
  }
  return "";
}


// ChopNick(NUH)
//
// Takes the nick out of a nick[user@host]
//
static std::string
ChopNick(std::string NUH)
{
  std::string::size_type pos;
  if (CountChars(NUH, '[') == 1) {
    pos = NUH.find('[');
    if (pos != std::string::npos)
      return NUH.substr(0, pos);
  }
  if (CountChars(NUH, '[') > 1) {
    if ((pos = NUH.find("[~")) != std::string::npos)
      return NUH.substr(0, pos);
    int nicklen = 0;
    while ((nicklen < 9) && (NUH.find('[', nicklen) != std::string::npos))
      nicklen++;
    return NUH.substr(0, nicklen - 1);
  }
  return "";
}


// Hash_Func(string)
//
//
//
static int
Hash_Func(const char *s)
{
  long i;

  i = *(s++);
  if (*s)
    i |= (*(s++) << 8);
    if (*s)
      i |= (*(s++) << 16);
      if (*s)
	i |= (*s << 24);
  return i % HASHTABLESIZE;
}

static int
Hash_IP_Func(unsigned long ip)
{
  return (ip & vhost_netmask) % HASHTABLESIZE;
}


// AddToHash(table[], key, item)
//
//
//
static void
AddToHash(HASHREC *table[], char *key, USERENTRY *item)
{
  int index;
  HASHREC *newhashrec;

  index = Hash_Func(key);
  newhashrec = (HASHREC *)malloc(sizeof(HASHREC));
  if (newhashrec == (HASHREC *)NULL) {
    gracefuldie(SIGTERM);
  }

  newhashrec->info = item;
  newhashrec->collision = table[index];
  table[index] = newhashrec;
}


// AddToIPHash(ip, item)
//
//
//
static void
AddToIPHash(unsigned long ip, USERENTRY *item)
{
  int index;
  HASHREC *newhashrec;

  index = Hash_IP_Func(ip);
  newhashrec = (HASHREC *)malloc(sizeof(HASHREC));
  if (newhashrec == (HASHREC *)NULL) {
    gracefuldie(SIGTERM);
  }

  newhashrec->info = item;
  newhashrec->collision = iptable[index];
  iptable[index] = newhashrec;
}


// RemoveFromHash(table[], key, hostmatch, usermatch, nickmatch)
//
//
//
static char
RemoveFromHash(HASHREC *table[], const char *key, const char *hostmatch,
  const char *usermatch, const char *nickmatch)
{
  int index;
  HASHREC *find, *prev;

  index = Hash_Func(key);
  find = table[index];
  prev = (HASHREC *)NULL;

  while (find) {
    if ((!hostmatch || !strcmp(find->info->host, hostmatch)) &&
	(!usermatch || !strcmp(find->info->user, usermatch)) &&
	(!nickmatch || !strcmp(find->info->nick, nickmatch))) {
      if (prev)
	prev->collision = find->collision;
      else
	table[index] = find->collision;

      if (find->info->link_count > 0)
	find->info->link_count--;

      if (find->info->link_count == 0) {
	(void)free(find->info);
      }
      (void)free(find);
      return 1;
    }
    prev = find;
    find = find->collision;
  }
  return 0;
}


// RemoveFromIPHash(ip, hostmatch, usermatch, nickmatch)
//
//
//
static char
RemoveFromIPHash(unsigned long ip, const char *hostmatch,
  const char *usermatch, const char *nickmatch)
{
  int index;
  HASHREC *find, *prev;

  index = Hash_IP_Func(ip);
  find = iptable[index];
  prev = (HASHREC *)NULL;

  while (find) {
    if ((!hostmatch || !strcmp(find->info->host, hostmatch)) &&
	(!usermatch || !strcmp(find->info->user, usermatch)) &&
	(!nickmatch || !strcmp(find->info->nick, nickmatch))) {
      if (prev)
	prev->collision = find->collision;
      else
	iptable[index] = find->collision;

      if (find->info->link_count > 0)
	find->info->link_count--;

      if (find->info->link_count == 0) {
	(void)free(find->info);
      }
      (void)free(find);
      return 1;
    }
    prev = find;
    find = find->collision;
  }
  // Didn't find it, try it without an IP
  if (ip)
    RemoveFromIPHash(0, hostmatch, usermatch, nickmatch);
  return 0;
}


// AddOper(notice)
//
//
//
void Add_Oper(std::string Notice)
{
}


// UpdateHash(table[], key, nick1, nick2)
//
//
//
static void
UpdateHash(HASHREC *table[], char *key, char *nick1, char *nick2)
{
  int index;
  HASHREC *find, *prev;

  index = Hash_Func(DownCase(key).c_str());
  find = table[index];

  while (find) {
    if (strcmp(find->info->nick, nick1) == 0)
    {
      strncpy(find->info->nick, nick2, MAX_NICK);
      find->info->nick[MAX_NICK - 1] = 0;
    }
    find = find->collision;
  }
}


// UpdateUserHost(nick1, nick2, userhost)
//
//
//
static void
UpdateUserHost(char *nick1, char *nick2, char *userhost)
{
  char *tmp, iphold[20];
  char *domain;
  char founddot = 0;
  int i = 0;

  tmp = strchr(userhost, '@');
  if (!tmp)
    return;
  *(tmp++) = 0;

  domain = tmp + strlen(tmp) - 1;
  if (isdigit(*domain)) {
    // IP
    domain = tmp;
    while (*domain) {
      iphold[i++] = *domain;
      if (*(domain++) == '.')
	founddot++;
      if (founddot == 3)
	break;

      if (*(domain++) == '.' && founddot++)
	break;
    }
    iphold[i] = '\0';
    domain = iphold;
 } else
   // FQDN
   while (--domain != tmp)
     if (*domain == '.' && founddot++)
       break;

   UpdateHash(hosttable, tmp, nick1, nick2);
}


// RemoveUserHost(nick, userhost, ip)
//
//
//
static void
RemoveUserHost(const std::string & nick, char *userhost, unsigned long ip)
{
  char *tmp, iphold[20];
  char *domain;
  char founddot = 0;
  int i = 0;

  tmp = strchr(userhost, '@');
  if (!tmp)
    return;
  *(tmp++) = 0;

  domain = tmp + strlen(tmp) - 1;
  if (isdigit(*domain)) {
    // IP
    domain = tmp;
    while (*domain) {
      iphold[i++] = *domain;
      if (*(domain++) == '.')
	founddot++;
      if (founddot == 3)
	break;
    }
    iphold[i] = '\0';
    domain = iphold;
  } else
    // FQDN
    while (--domain != tmp)
      if (*domain == '.' && founddot++)
	break;

  if (!RemoveFromHash(hosttable, tmp, tmp, userhost, nick.c_str()))
    if (!RemoveFromHash(hosttable, tmp, tmp, userhost, NULL))
      Log::Write("Error removing user from host table!");

  if (!RemoveFromHash(domaintable, domain+1, tmp, userhost, nick.c_str()))
    if (!RemoveFromHash(domaintable, domain+1, tmp, userhost, NULL))
      Log::Write("Error removing user from domain table!");

  if (!RemoveFromHash(usertable, userhost, tmp, userhost, nick.c_str()))
    if (!RemoveFromHash(usertable, userhost, tmp, userhost, NULL))
      Log::Write("Error removing user from user table!");

  if (!RemoveFromIPHash(ip, tmp, userhost, nick.c_str()))
    if (!RemoveFromIPHash(ip, tmp, userhost, NULL))
      Log::Write("Error removing user from ip table!");
}


// AddUserHost(Nick, UserHost, IP, FromTrace, IsOper, Class)
//
//
//
static void
AddUserHost(const std::string & nick, char *userhost,
  const std::string & ip, bool fromtrace, bool isoper,
  const std::string & Class)
{
  USERENTRY *newuser;
  char *host, iphold[20];
  char hold[MAX_BUFF];
  char *domain;
  char founddot = 0;
  int i = 0;
  char *p;
  char *user = userhost;
  std::string UserHost = userhost;

  host = strrchr(userhost, '@');
  if (!host)
    return;
  *(host++) = '\0';

  p = strrchr(host, '.');
  if (p) {
    int len;
    bool legal_top_level = false;

    p++;
    len = strlen(p);
    if (len > 3) {
      // Don't forget .ARPA !!!! :P
      if (strcasecmp(p, "arpa")==0) legal_top_level = true;
#ifdef KILL_ILLEGAL_TLD_DNS_SPOOFERS
      if (!legal_top_level && !Config::IsSpoofer(ip)) {
        Kill_Spoof(nick, UserHost, ip);
        return;
      }
#endif
    }

    if (len == 3) {
      if (strcasecmp(p, "net")==0) legal_top_level = true;
      if (strcasecmp(p, "com")==0) legal_top_level = true;
      if (strcasecmp(p, "org")==0) legal_top_level = true;
      if (strcasecmp(p, "gov")==0) legal_top_level = true;
      if (strcasecmp(p, "edu")==0) legal_top_level = true;
      if (strcasecmp(p, "mil")==0) legal_top_level = true;
      if (strcasecmp(p, "int")==0) legal_top_level = true;

      if (isdigit(*p) && isdigit(*(p+1)) && isdigit(*(p+2)))
	legal_top_level = true;

#ifdef KILL_ILLEGAL_TLD_DNS_SPOOFERS
      if (!legal_top_level && !Config::IsSpoofer(ip)) {
	Kill_Spoof(nick, UserHost, ip);
	return;
      }
#endif
    }
  }

#ifdef KILL_ILLEGAL_CHAR_SPOOFERS
  if (strchr(host, '@') && !Config::IsSpoofer(ip)) {
    Kill_Spoof(nick, UserHost, ip);
    return;
  }

  if (strchr(host, '*') && !Config::IsSpoofer(ip)) {
    Kill_Spoof(nick, UserHost, ip);
    return;
  }

  if (strchr(host, '?') && !Config::IsSpoofer(ip)) {
    Kill_Spoof(nick, UserHost, ip);
    return;
  }
#endif

  domain = host + strlen(host) - 1;

#ifdef KILL_FAKE_IP_SPOOFERS
  // If we're dealing with an IP that that doesn't reverse-resolve,
  // make sure the "hostname" and ip match up.
  if ((ip != "") && isdigit(*domain)) {
    if (strcmp(host, ip.c_str()) && !Config::IsSpoofer(ip)) {
      Kill_Spoof(nick, UserHost, ip);
      return;
    }
  }
#endif

  // If we made it this far, we'll just assume it isn't a spoofed
  // hostname.  Oh, happy day!

  newuser = (USERENTRY *)malloc(sizeof(USERENTRY));
  if (newuser == (USERENTRY *)NULL) {
    gracefuldie(SIGTERM);
  }

  strncpy(newuser->nick, nick.c_str(), MAX_NICK);
  newuser->nick[MAX_NICK-1] = '\0';
  strncpy(newuser->user, user, 11);
  newuser->user[10] = '\0';
  strncpy(newuser->host, host, MAX_HOST);
  newuser->host[MAX_HOST-1] = '\0';
  if (ip != "")
    newuser->ip = inet_addr(ip.c_str());
  newuser->connecttime = (fromtrace ? 0 : time(NULL));
  newuser->reporttime = 0;
  newuser->link_count = 4;
  newuser->isoper = isoper;
  strncpy(newuser->userclass, Class.c_str(), MAX_CLASS);
  newuser->userclass[MAX_CLASS - 1] = '\0';

  // Determine the domain name type
  if (isdigit(*domain)) {
    // IP
    domain = host;
    while (*domain) {
      iphold[i++] = *domain;
      if (*(domain++) == '.')
	founddot++;
      if (founddot == 3)
	break;
    }
    iphold[i] = '\0';
    domain = iphold;
  } else
    while (--domain != host)
      if (*domain == '.' && founddot++)
	break;
  strncpy(newuser->domain, (*domain=='.' ? domain+1 : domain), MAX_DOMAIN);
  newuser->domain[MAX_DOMAIN-1] = '\0';

  // Add it to the hash tables
  AddToHash(hosttable, host, newuser);
  AddToHash(domaintable, domain+1, newuser);
  AddToHash(usertable, user, newuser);
  AddToIPHash(newuser->ip, newuser);

  // Don't check for wingate or clones when doing a TRACE
  if (!fromtrace)
  {
    if (!Config::IsOKHost(UserHost, ip) && !Config::IsOper(UserHost, ip))
    {
      // Check if this client matches any traps
      TrapList::match(nick, UserHost, ip);

      if (ip != "")
      {
        CheckProxy(ip, host, nick, UserHost);

#ifdef DETECT_VHOST_CLONES
        Add_IP(ip.c_str());
#endif
      }

      // Clonebot check
      CheckHostClones(host);
    }
  }
}


// OnTraceUser(Type, Class, Nick, UserHost, IP)
//
//
//
void
OnTraceUser(const std::string & Type, const std::string & Class,
  const std::string & Nick, std::string UserHost,
  std::string IP = "")
{
  char userhost[512];

  if (!DoingTrace || (Type == "") || (Nick == "") || (UserHost == ""))
    return;

  // Remove the brackets surrounding the user@host
  if (UserHost.length() >= 2)
  {
    UserHost = UserHost.substr(1, UserHost.length() - 2);
  }

  // This might not be necessary, but we'll do it just in case...
  if ((UserHost.length() >= 3) && (UserHost.substr(0, 3) == "(+)"))
  {
    UserHost = UserHost.substr(3, std::string::npos);
  }

  // Copy UserHost to a char array and null terminate
  strncpy(userhost, UserHost.c_str(), sizeof(userhost) - 1);
  userhost[sizeof(userhost) - 1] = 0;

  // Remove the parentheses surrounding the IP
  if (IP.length() >= 2)
  {
    IP = IP.substr(1, IP.length() - 2);
  }

  AddUserHost(Nick, userhost, IP, true, (Type == "Oper"), Class);
}


// OnTraceUser(Type, Class, NUH, IP)
//
//
//
void OnTraceUser(const std::string & Type, const std::string & Class,
  const std::string & NUH, std::string IP = "")
{
  std::string Nick;
  std::string UserHost;
  char userhost[512];

  if (!DoingTrace || (Type == "") || (NUH == ""))
    return;

  // Split the nick and user@host from the nick[user@host]
  Nick = ChopNick(NUH);
  UserHost = DownCase(ChopUH(NUH));

  // hybrid-6 borked trace :P
  if ((UserHost.length() >= 3) && (UserHost.substr(0, 3) == "(+)"))
  {
    UserHost = UserHost.substr(3, std::string::npos);
  }

  // Copy UserHost to a char array and null terminate
  strncpy(userhost, UserHost.c_str(), sizeof(userhost) - 1);
  userhost[sizeof(userhost) - 1] = 0;

  // Remove the parentheses surrounding the IP
  if (IP.length() >= 2)
  {
    IP = IP.substr(1, IP.length() - 2);
  }

  AddUserHost(Nick, userhost, IP, true, (Type == "Oper"), Class);
}

void OnTraceClass()
{
  if (DoingTrace)
  {
    DoingTrace = false;
    ::Status("TRACE complete.", SEND_OPERS_ONLY);
  }
}

void List_Users(StrList & Output, const std::string & UserHost)
{
  HASHREC *userptr;
  int i, numfound = 0;
  char outmsg[MAX_BUFF];

  Output.clear();

  if ((UserHost == "*") || (UserHost == "*@*"))
  {
    Output.push_back("Listing all users is not recommended. To do it anyway,");
    Output.push_back("use 'list ?*@*'.");
  }
  else
  {
    for (i = 0; i < HASHTABLESIZE; ++i)
    {
      userptr = domaintable[i];
      while (userptr)
      {
	std::string fullUH = std::string(userptr->info->user) + "@" +
	  std::string(userptr->info->host);
	bool matchesUH = MatchesMask(fullUH, UserHost);
	bool matchesUIP = false;
	if (userptr->info->ip)
	{
	  struct in_addr in;
	  in.s_addr = userptr->info->ip;
	  std::string fullUIP = std::string(userptr->info->user) + "@" +
	    std::string(inet_ntoa(in));
	  matchesUIP = MatchesMask(fullUIP, UserHost);
	}
	if (matchesUH || matchesUIP)
	{
	  if (!numfound++)
	  {
	    Output.push_back("The following clients match " + UserHost + ":");
	  }
	  if (userptr->info->ip)
	  {
	    struct in_addr in;
	    in.s_addr = userptr->info->ip;
	    sprintf(outmsg, "  %s (%s) [%s]", userptr->info->nick,
	      fullUH.c_str(), inet_ntoa(in));
	  }
	  else
	  {
	    sprintf(outmsg, "  %s (%s)", userptr->info->nick, fullUH.c_str());
	  }
	  Output.push_back(outmsg);
	}
	userptr = userptr->collision;
      }
    }
    if (numfound > 0)
    {
      sprintf(outmsg, "%d matches for %s found", numfound,
        UserHost.c_str());
    }
    else
    {
      sprintf(outmsg, "No matches for %s found", UserHost.c_str());
    }
    Output.push_back(outmsg);
  }
}

void OnClientConnect(std::string Text)
{
// IRC >> :plasma.toast.pc NOTICE OOMon :*** Notice -- Client connecting: Toast (toast@Plasma.Toast.PC) [192.168.1.1] {opers} [http://toast.blackened.com/]
  std::string Nick = FirstWord(Text);
  std::string UH = FirstWord(Text);
  std::string IP = FirstWord(Text);
  std::string Class = FirstWord(Text);
  if ((UH[0] == '(') && (UH[UH.length() - 1] == ')'))
  {
    char userhost[512];
  
    // Remove the parentheses surrounding the user@host and copy to a char
    // array for later disposal.
    if (UH.length() >= 2)
    {
      UH = UH.substr(1, UH.length() - 2);
    }
    strncpy(userhost, DownCase(UH).c_str(), sizeof(userhost) - 1);
    userhost[sizeof(userhost) - 1] = 0;

    // Remove the brackets surrounding the IP
    if (IP.length() >= 2)
    {
      IP = IP.substr(1, IP.length() - 2);
    }

    // Remove the curly brackets surrounding the class
    if (Class.length() >= 2)
    {
      Class = Class.substr(1, Class.length() - 2);
    }

    AddUserHost(Nick, userhost, IP, false, false, Class);
  }
}

void OnClientExit(std::string Text)
{
  // "cihuyy (ngocol@p6.nas2.is4.u-net.net) [Leaving] [195.102.196.13]"
  std::string Nick = FirstWord(Text);
  std::string UH = FirstWord(Text);
  if (UH.length() >= 2)
  {
    std::string IP;
    char userhost[512];
    unsigned long ip;
    std::string::size_type i;
  
    UH = UH.substr(1, UH.length() - 2);
    strncpy(userhost, DownCase(UH).c_str(), sizeof(userhost) - 1);
    userhost[sizeof(userhost) - 1] = 0;

    if ((i = Text.rfind(' ')) != std::string::npos) {
      IP = Text.substr(i + 1);
      IP = IP.substr(1, IP.length() - 2);
      ip = inet_addr(IP.c_str());
    } else
      ip = 0;

    RemoveUserHost(Nick.c_str(), userhost, ip);
  }
}

void Bot_Reject(std::string Text)
{
}

void CS_Clones(std::string Text)
{
  bool identd = true;
  char server_notice[MAX_BUFF];
  char *user;
  char *host;
  char *p;
  char *nick_reported;
  char *user_host; 
  char notice[MAX_BUFF];

  strncpy(server_notice, Text.c_str(), sizeof(server_notice));
  server_notice[sizeof(server_notice) - 1] = 0;

  nick_reported = strtok(server_notice," ");
  if(nick_reported == (char *)NULL)return;
    
  user_host = strtok((char *)NULL," ");
  if(user_host == (char *)NULL)return;
    
/*
  Lets try and get it right folks... [user@host] or (user@host)
*/  
 
  if(*user_host == '[')
    {
      user_host++;
      p = strrchr(user_host,']');
      if(p)  
        *p = '\0';
    }
  else if(*user_host == '(')
    {
      user_host++;
      p = strrchr(user_host,')');
      if(p)
        *p = '\0'; 
    }
  
  ::Status("CS clones: " + std::string(user_host), SEND_OPERS_ONLY);
  Log::Write("CS clones: " + std::string(user_host));
  
  p = strdup(user_host);
  user = strtok(user_host,"@");
  if(*user == '~')
    {
      user++;
      identd = false;
    } 
      
  host = strtok((char*)NULL,"");
    
  Suggest_Kline(false,user,host,false,identd,MK_CLONES);
  
  (void)free(p);
}     

void CS_Nick_Flood(std::string Text)
{
  char server_notice[MAX_BUFF];
  char *nick_reported;
  char *user_host;
  char *user;
  char *host;
  char *p;
  char notice[MAX_BUFF];

  strncpy(server_notice, Text.c_str(), sizeof(server_notice));
  server_notice[sizeof(server_notice) - 1] = 0;

  nick_reported = strtok(server_notice," ");
  if(nick_reported == (char *)NULL)return;
      
  user_host = strtok((char *)NULL," ");
  if(user_host == (char *)NULL)return;

/*    
  Lets try and get it right folks... [user@host] or (user@host)
*/      
      
  if(*user_host == '[')
    {
      user_host++;
      p = strrchr(user_host,']');
      if(p)
        *p = '\0';
    }
  else if(*user_host == '(')
    {
      user_host++;
      p = strrchr(user_host,')');
      if(p)
        *p = '\0'; 
    } 
  
    ::Status("CS nick flood: " + std::string(user_host), SEND_OPERS_ONLY);
    Log::Write("CS nick flood: " + std::string(user_host));
      
  if ( (!Config::IsOKHost(user_host)) && (!Config::IsOper(user_host)) )
    {   
#ifdef AUTO_KILL_NICK_FLOODING
      if(autopilot)
        {
          IRC::Kill("Auto-Kill", nick_reported, "flooding");
        }  
#endif  
      user = strtok(user_host,"@");
      if(user == (char *)NULL)return; /* OOOPSIES! */
      host = strtok((char *)NULL,"");
      if(host == (char *)NULL)return; /* OOPSIES! */
      
#ifdef AUTO_KILL_NICK_FLOODING
/*
   IF AUTO killing nick flooders, I will not be k-lining
   nick flooders
*/    
      if(*user_host == '~')
        Suggest_Kline(false,user,host,false,false,MK_FLOODING);
      else
        Suggest_Kline(false,user,host,false,true,MK_FLOODING);
#else
/*    
   IF not AUTO killing nick flooders, might still be AUTO k-lining
   nick flooders
*/        
      if(*user_host == '~')
        Suggest_Kline(true,user,host,false,false,MK_FLOODING);
      else 
        Suggest_Kline(true,user,host,false,true,MK_FLOODING);
#endif
    } 
}     

void Check_Nick_Flood(std::string Text)
{
  char server_notice[1024];
  char *p;
  char *nick1;  
  char *nick2;
  char *user_host;

  strncpy(server_notice, Text.c_str(), sizeof(server_notice));
  server_notice[sizeof(server_notice) - 1] = 0;
      
#ifdef TELETYPE_NIH
  nick1 = strtok(server_notice," ");    /* This _should_ be nick1 */
  if(nick1 == (char *)NULL)return;
#else
  p = strtok(server_notice," ");        /* Throw away the "From" */
  if(p == (char *)NULL) return;
      
  if(strcasecmp(p,"From") != 0) /* This isn't an LT notice */
    {
      nick1 = p;        /* This _should_ be nick1 */
      
      user_host = strtok((char *)NULL," ");     /* (user@host) */
      if(user_host == (char *)NULL)return;
      if(*user_host == '(')
        user_host++;
        
      p = strrchr(user_host,')');
      if(p)
        *p = '\0';
  
      p = strtok((char *)NULL," ");
      if(p == (char *)NULL)
        return;
  
      if(strcmp(p,"now") != 0 )
        return;
  
      p = strtok((char *)NULL," ");
      if(p == (char *)NULL)
        return;
      
      if(strcmp(p,"known") != 0 )
        return; 
      
      p = strtok((char *)NULL," ");
      if(p == (char *)NULL)
        return;
      
      if(strcmp(p,"as") != 0 )
        return;
  
      nick2 = strtok((char *)NULL," ");/* This _should_ be nick2 */
      if(nick2 == (char *)NULL)return;
        
      Add_To_Nick_Change_Table(user_host,nick2);
      UpdateUserHost(nick1,nick2,user_host);
        
      return;
    } 
      
  nick1 = strtok((char *)NULL," ");     /* This _should_ be nick1 */
  if(nick1 == (char *)NULL)return;
#endif
        
  p = strtok((char *)NULL," ");         /* Throw away the "to" */
  if(p == (char *)NULL)return;
      
  nick2 = strtok((char *)NULL," ");     /* This _should_ be nick2 */
  if(nick2 == (char *)NULL)return;
      
  user_host = strtok((char *)NULL," "); /* and this _should_ be the u@h part */
  if(user_host == (char *)NULL)return;
      
  if(*user_host == '[') 
    user_host++;
  
  p = strrchr(user_host,']');
  if(p) 
    *p = '\0';
    
/* N.B.
   hendrix's original code munges the user_host variable 
   so, add_to_nick_change must occur BEFORE
   updateuserhost is called. grrrrrrrrrrrr
   I hate order dependencies of calls.. but there you are.
   This caused a bug in v0.1
  
   -Dianora
*/
  
  Add_To_Nick_Change_Table(user_host,nick2);
  UpdateUserHost(nick1,nick2,user_host);
}

struct sortarray sort[MAXDOMAINS];
void Report_Domains(StrList & Output, const int num)
{
  struct hashrec *userptr;          
             
  int inuse = 0;
  int i,j,maxx,found,foundany = 0;
  char outmsg[MAX_BUFF];            

  Output.clear();

  for (i=0;i<HASHTABLESIZE;++i) {
    userptr = hosttable[i];
    while (userptr) {   
      for (j=0;j<inuse;++j)
        if (!strcasecmp(userptr->info->domain,sort[j].domainrec->domain))
          break;
      if (j == inuse && inuse < MAXDOMAINS) {
        sort[inuse].domainrec = userptr->info;
        sort[inuse++].count = 1;
        }
      else
        ++sort[j].count;
      userptr = userptr->collision;
      }
    }
  /* Print 'em out from highest to lowest */
  for (;;) { 
    maxx = num-1;
    found = -1;
    for (i=0;i<inuse;++i)           
      if (sort[i].count > maxx) {
        found = i;
        maxx = sort[i].count;
        } 
    if (found == -1)
      break;
    if (!foundany++)
      Output.push_back("Domains with most users on the server:");
    (void)sprintf(outmsg,
                  "  %-40s %3d users",sort[found].domainrec->domain,maxx);
    Output.push_back(outmsg);
    sort[found].count = 0;
    }   
  if (!foundany)
    {  
      (void)sprintf(outmsg,
                    "No domains have %d or more users.",num);
      Output.push_back(outmsg);
    }
}

void Link_Look_Notice(std::string Text)
{
  char server_notice[MAX_BUFF];
  char *nick_reported;
  char *user_host;
  char *user;
  char *host;
  char *p;
  char notice[MAX_BUFF];
  char hold[MAX_BUFF];
  char copy_of_server_notice[MAX_BUFF];
  time_t current_time;
  int first_empty_entry = -1;
  bool found_entry = false;
  int i;

  strncpy(server_notice, Text.c_str(), sizeof(server_notice));
  server_notice[sizeof(server_notice) - 1] = 0;

  current_time = time((time_t *)NULL);

  p = server_notice;

  if(*p == '\'')                /* Is this a +th server ?*/
  {
    /* This is a +th server, skip the '...' part */
    p++;
    p = strchr(p,'\'');
    if(p == (char*)NULL)return;       /* just ignore it *sigh* */
    p++;
    strncpy(copy_of_server_notice, p, sizeof(copy_of_server_notice));
    copy_of_server_notice[sizeof(copy_of_server_notice) - 1] = 0;
  }
  else 
  {
    strncpy(copy_of_server_notice, p, sizeof(copy_of_server_notice));
    copy_of_server_notice[sizeof(copy_of_server_notice) - 1] = 0;
  }
  
  p = strtok(copy_of_server_notice," ");
  if(p == (char *)NULL) return; /* just ignore it *sigh* */
  
  if(strcasecmp(p,"requested") != 0)return;     /* just ignore it *sigh* */
  
  p = strtok((char *)NULL," ");
  if(p == (char *)NULL) return;         /* just ignore it *sigh* */
  if(strcasecmp(p,"by") != 0)return;    /* just ignore it *sigh* */
  
  nick_reported = strtok((char *)NULL," ");
  if(nick_reported == (char *)NULL)return;
      
  user_host = strtok((char *)NULL," ");
  if(user_host == (char *)NULL)return;  
      
/*
  Lets try and get it right folks... [user@host] or (user@host)
*/
  
  if (*user_host == '[') {
    user_host++;
    p = strrchr(user_host,']'); 
    if (p)
      *p = '\0'; 
  } else if (*user_host == '(') {
    user_host++;
    p = strrchr(user_host,')');
    if (p)
      *p = '\0';
  }
      
  /* Don't even complain about opers */
  
  if (Config::IsOper(user_host))
    return;
      
  ::Status("[LINKS " + Text + "]", SEND_OPERS_ONLY);
    
  for(i = 0; i < LINK_LOOK_TABLE_SIZE; i++ )
    {
      if(link_look[i].user_host)
        { 
          if(strcasecmp(link_look[i].user_host,user_host) == 0)
            {
              char *user;
              char *host;
  
              found_entry = true;
  
              /* if its an old old entry, let it drop to 0, then start counting
                 (this should be very unlikely case)
                 */
              if((link_look[i].last_link_look + MAX_LINK_TIME) < current_time)
                {
                  link_look[i].link_look_count = 0;
                }
  
              link_look[i].link_look_count++;
      
              if(link_look[i].link_look_count >= MAX_LINK_LOOKS)
                {
                  ::Status("Possible links looker: " +
		    std::string(nick_reported) + " (" +
		    std::string(user_host) + ")", SEND_OPERS_ONLY);
		  Log::Write("Possible links looker: " + std::string(nick_reported) +
		    " (" + std::string(user_host) + ")");
  
/*              
   The code as it is, doesn't AUTO_KLINE link lookers ever, but will
   AUTO_KILL link lookers IFF autopilot is on. HOWEVER,
   Shadowfax still wants to see a suggested kline to use.
     
   - Dianora
*/            
                  if ( !Config::IsOKHost(user_host) )
                    {
#ifdef AUTO_KILL_LINK_LOOKERS   
                      if(autopilot)
                        {
                          IRC::Kill("Auto-Kill", nick_reported, "links looker");
                          (void)free(link_look[i].user_host);
                          link_look[i].user_host = (char *)NULL;
                        }     
#endif                        
                      user = strtok(user_host,"@");
                      if(user == (char *)NULL)return; /* OOOPSIES! */
                      host = strtok((char *)NULL,"");
                      if(host == (char *)NULL)return; /* OOPSIES! */
   
                      if(*user_host == '~') 
                        Suggest_Kline(false,user,host,false,false,MK_LINKS);
                      else
                        Suggest_Kline(false,user,host,false,true,MK_LINKS);
                    }
                    
#ifndef AUTO_KILL_LINK_LOOKERS  
                  link_look[i].last_link_look = current_time;
#endif
                }       
              else            
                {     
                  link_look[i].last_link_look = current_time;
                }     
            }         
          else
            {         
              if((link_look[i].last_link_look + MAX_LINK_TIME) < current_time)
                {     
                  (void)free(link_look[i].user_host);
                  link_look[i].user_host = (char *)NULL;
                }   
            }
        }         
      else
        {         
          if(first_empty_entry < 0)
            first_empty_entry = i;
        }             
    }               
                
/*            
   If this is a new entry, then found_entry will still be NO
*/                
                
  if(!found_entry)    
    {     
      if(first_empty_entry >= 0)
        {     
          link_look[first_empty_entry].user_host = strdup(user_host);
          link_look[first_empty_entry].last_link_look = current_time;
          link_look[first_empty_entry].link_look_count = 0;
        }   
    }   
}

void Report_Clones(StrList & Output)
{
  struct hashrec *userptr, *top, *temp;
  int numfound, i, j, k;
  char notip, foundany = 0;
  char outmsg[MAX_BUFF];
  time_t connfromhost[MAXFROMHOST];

  Output.clear();

  for (i = 0; i < HASHTABLESIZE; ++i) {
    top = userptr = hosttable[i];
    while (userptr) {
      numfound = 0;
      temp = top;
      while (temp != userptr)
        if (!strcmp(temp->info->host, userptr->info->host))
	  break;
        else
	  temp = temp->collision;
      if (temp == userptr) {
        connfromhost[numfound++] = temp->info->connecttime;
        temp = temp->collision;
        while (temp) {
          if (!strcmp(temp->info->host,userptr->info->host) &&
              numfound < MAXFROMHOST)
            connfromhost[numfound++] = temp->info->connecttime;
          temp = temp->collision;
          }
        if (numfound > 2) {
          for (k=numfound-1;k>1;--k)
            for (j=0;j<numfound-k;++j) {
              if (connfromhost[j] &&
                  connfromhost[j] - connfromhost[j+k] <= (k+1) * CLONEDETECTINC)
                goto getout;  /* goto rules! */
              }
          getout:
          if (k > 1) {
            if (!foundany++)
              Output.push_back("Possible clonebots from the following hosts:");
            sprintf(outmsg,
                    "  %2d connections in %3d seconds (%2d total) from %s",
                    k+1, connfromhost[j] - connfromhost[j+k], numfound+1,
                    userptr->info->host);
            Output.push_back(outmsg);
          }   
        } 
      }     
      userptr = userptr->collision;
    }
  }
  if (!foundany)
    Output.push_back("No potential clonebots found.");
}

void Report_Multi(StrList & Output, const int Minimum)
{
  struct hashrec *userptr,*top,*temp;
  int numfound,i;
  char notip, foundany = 0;
  char outmsg[MAX_BUFF];
  int minclones = MULTI_MINIMUM;

  if (Minimum > 0)
    minclones = Minimum;

  Output.clear();

  for (i=0;i<HASHTABLESIZE;++i) {
    top = userptr = domaintable[i];
    while (userptr) {
      numfound = 0;
      /* Ensure we haven't already checked this user & domain */
      temp = top;
      while (temp != userptr)
        if (!strcmp(temp->info->user,userptr->info->user) &&
            !strcmp(temp->info->domain,userptr->info->domain))
          break;
        else
          temp = temp->collision;
      if (temp == userptr) {
        temp = temp->collision;
        while (temp) {
          if (!strcmp(temp->info->user,userptr->info->user) &&
              !strcmp(temp->info->domain,userptr->info->domain))
#ifdef NO_OPERS_IN_MULTI
            if (!temp->info->isoper &&
	      !Config::IsOper(std::string(temp->info->user) + "@" +
	      std::string(temp->info->host)))
#endif
              numfound++;       /* - zaph & Dianora :-) */
          temp = temp->collision;
        }
        if (numfound >= (minclones - 1))
	{
          if (!foundany++)
            Output.push_back("Multiple clients from the following userhosts:");
          notip = strncmp(userptr->info->domain, userptr->info->host,
            strlen(userptr->info->domain)) ||
            (strlen(userptr->info->domain) == strlen(userptr->info->host));
          numfound++; /* - zaph and next line*/
          sprintf(outmsg, " %s %2d -- %s@%s%s",
            (numfound > minclones) ? "==>" : "   ", numfound,
	    userptr->info->user, notip ? "*" : userptr->info->domain,
	    notip ? userptr->info->domain : "*", numfound);
          Output.push_back(outmsg);
        }
      }  
      userptr = userptr->collision;
    }
  }         
  if (!foundany) 
    Output.push_back("No multiple logins found.");
}

void Report_HMulti(StrList & Output, const int Minimum)
{
  struct hashrec *userptr,*top,*temp;
  int numfound,i;
  char foundany = 0;
  char outmsg[MAX_BUFF];
  int minclones = MULTI_MINIMUM;

  if (Minimum > 0)
    minclones = Minimum;

  Output.clear();

  for (i=0;i<HASHTABLESIZE;++i) {
    top = userptr = domaintable[i];
    while (userptr) {
      numfound = 0;
      // Ensure we haven't already checked this hostname
      temp = top;
      while (temp != userptr)
        if (!strcmp(temp->info->host,userptr->info->host))
          break;
        else
          temp = temp->collision;
      if (temp == userptr) {
        temp = temp->collision;
        while (temp) {
          if (!strcmp(temp->info->host,userptr->info->host))
#ifdef NO_OPERS_IN_MULTI
            if (!temp->info->isoper &&
	      !Config::IsOper(std::string(temp->info->user) + "@" +
	      std::string(temp->info->host)))
#endif
              numfound++;       /* - zaph & Dianora :-) */
          temp = temp->collision;
        }
        if (numfound >= (minclones - 1))
	{
          if (!foundany++)
            Output.push_back("Multiple clients from the following hosts:");
          numfound++; /* - zaph and next line*/
          sprintf(outmsg, " %s %2d -- *@%s",
	    (numfound > minclones) ? "==>" : "   ",numfound,
	    userptr->info->host);
          Output.push_back(outmsg);
        }
      }  
      userptr = userptr->collision;
    }
  }         
  if (!foundany) 
    Output.push_back("No multiple logins found.");
}

void Report_UMulti(StrList & Output, const int Minimum)
{
  struct hashrec *userptr,*top,*temp;
  int numfound,i;
  char foundany = 0;
  char outmsg[MAX_BUFF];
  int minclones = MULTI_MINIMUM;

  if (Minimum > 0)
    minclones = Minimum;

  Output.clear();

  for (i=0;i<HASHTABLESIZE;++i) {
    top = userptr = usertable[i];
    while (userptr) {
      numfound = 0;
      // Ensure we haven't already checked this username
      temp = top;
      while (temp != userptr)
        if (!strcmp(temp->info->user,userptr->info->user))
          break;
        else
          temp = temp->collision;
      if (temp == userptr) {
        temp = temp->collision;
        while (temp) {
          if (!strcmp(temp->info->user,userptr->info->user))
#ifdef NO_OPERS_IN_MULTI
            if (!temp->info->isoper &&
	      !Config::IsOper(std::string(temp->info->user) + "@" +
	      std::string(temp->info->host)))
#endif
              numfound++;       /* - zaph & Dianora :-) */
          temp = temp->collision;
        }
        if (numfound >= (minclones - 1))
	{
          if (!foundany++)
            Output.push_back("Multiple clients from the following usernames:");
          numfound++; /* - zaph and next line*/
          sprintf(outmsg, " %s %2d -- %s@*",
	    (numfound > minclones) ? "==>" : "   ", numfound,
	    userptr->info->user);
          Output.push_back(outmsg);
        }
      }  
      userptr = userptr->collision;
    }
  }         
  if (!foundany) 
    Output.push_back("No multiple logins found.");
}

void Report_VMulti(StrList & Output, const int Minimum)
{
  struct hashrec *userptr,*top,*temp;
  int numfound,i;
  char notip, foundany = 0;
  char outmsg[MAX_BUFF];
  int minclones = MULTI_MINIMUM;
  struct in_addr in;
  char *IP;
  char *dot;

  if (Minimum > 0)
    minclones = Minimum;

  Output.clear();

  for (i=0;i<HASHTABLESIZE;++i) {
    top = userptr = iptable[i];
    while (userptr) {
      numfound = 0;
      /* Ensure we haven't already checked this user & domain */
      temp = top;
      while (temp != userptr)
        if (!strcmp(temp->info->user,userptr->info->user) &&
            ((userptr->info->ip & vhost_netmask) == (userptr->info->ip & vhost_netmask)))
          break;
        else
          temp = temp->collision;
      if (temp == userptr) {
        temp = temp->collision;
        while (temp) {
          if (!strcmp(temp->info->user,userptr->info->user) &&
              ((userptr->info->ip & vhost_netmask) == (userptr->info->ip & vhost_netmask)))
#ifdef NO_OPERS_IN_MULTI
            if (!temp->info->isoper &&
	      !Config::IsOper(std::string(temp->info->user) + "@" +
	      std::string(temp->info->host)))
#endif
              numfound++;       /* - zaph & Dianora :-) */
          temp = temp->collision;
        }
        if (numfound >= (minclones - 1))
	{
          if (!foundany++)
            Output.push_back("Multiple clients from the following vhosts:");
          numfound++; /* - zaph and next line*/
	  in.s_addr = userptr->info->ip;
	  IP = inet_ntoa(in);
          dot = rindex(IP, '.');
	  if (dot)
	  {
	    *dot = 0;
	  }
          sprintf(outmsg, " %s %2d -- %s@%s.*",
	    (numfound > minclones) ? "==>" : "   ", numfound,
	    userptr->info->user, IP);
          Output.push_back(outmsg);
        }
      }  
      userptr = userptr->collision;
    }
  }         
  if (!foundany) 
    Output.push_back("No multiple logins found.");
}

int
Kill_List_Users(const std::string & From, const std::string & UserHost,
  const std::string & Reason)
{
  int numfound = 0;

  Log::Write("KILLLIST " + UserHost + " (" + Reason + ") [" + From + "]");

  for (int i = 0; i < HASHTABLESIZE; ++i)
  {
    struct hashrec *userptr = domaintable[i];
    while (userptr)
    {
      std::string fullUH = std::string(userptr->info->user) + "@" +
	std::string(userptr->info->host);
      if (MatchesMask(fullUH, UserHost))
      {
	IRC::Kill(From, userptr->info->nick, Reason);
	numfound++;
      }
      userptr = userptr->collision;
    }
  } 
  return numfound;
}

int
Kill_List_Users(const std::string & From, const std::string & UserHost)
{
  return Kill_List_Users(From, UserHost, DEFAULT_KILLLIST_REASON);
}

bool FoundUser(const std::string & Nick)
{
  HASHREC *find;
      
  find = hosttable[0];  
  for (int i = 0; i < HASHTABLESIZE; ++i) {
    find = hosttable[i];
    while (find) {
      if (!strcasecmp(find->info->nick, Nick.c_str()))
        return true;
      find = find->collision;
    }
  }
  return false;
}

// "OOMon. From Toast Path: plasma.engr.Arizona.EDU!Toast (test)"
void Kill_Add_Report(std::string Text)
{
  bool Global;

  std::string Nick = FirstWord(Text); // Killed nick
  if (Nick[Nick.length() - 1] == '.') // Remove the trailing .
    Nick = Nick.substr(0, Nick.length() - 1);
  if (FoundUser(Nick)) {
    // We only want to report kills on local users
    FirstWord(Text); // "From"
    std::string Killer = FirstWord(Text); // Killer's nick
    if ((std::string::npos != Killer.find('@')) || 
      (std::string::npos != Killer.find('!')) || 
      (std::string::npos == Killer.find('.')))
    {
      // We don't want to see server kills :P
      FirstWord(Text); // "Path:"
      std::string Path = FirstWord(Text); // Path
      // Hybrid-6 changes the format of kill paths
      Global = !FoundUser(Killer);
      if (Global) {
        ::Status("Global kill for " + Nick + " by " + Killer + " " + Text,
	  SEND_OPERS_ONLY);
	Log::Write("Global kill for " + Nick + " by " + Killer + " " + Text);
      } else {
        ::Status("Local kill for " + Nick + " by " + Killer + " " + Text,
	  SEND_OPERS_ONLY);
	Log::Write("Local kill for " + Nick + " by " + Killer + " " + Text);
      }
    }
  }
}

void Stats_Notice(std::string Text)
{
  char temp[MAX_BUFF];
  char *server_notice;
  char *requested;
  char *by;
  char *nick;

  strncpy(temp, Text.c_str(), sizeof(temp));
  temp[sizeof(temp) - 1] = 0;
  server_notice = temp;
 
  if ((*server_notice == 'p')
#ifdef STATS_P_CASE_INSENSITIVE
    || (*server_notice == 'P')
#endif
    )
  {
    server_notice += 2;
    requested = strtok(server_notice," ");
    if(requested == (char *)NULL)
    {
      return;
    }

    if (strcasecmp(requested,"requested"))
    {
      return;
    }
    by = strtok((char *)NULL," ");
    if (by == (char *)NULL)
    {
      return;
    }
    if(strcasecmp(by,"by"))
    {
      return;
    }
    nick = strtok((char *)NULL," ");

    ::Status("[STATS " + Text + "]", SEND_OPERS_ONLY);

#ifdef SHOW_STATSP
    // Assume this is a server with stats P support.  Only report
    // bot DCC connections with +O status
    StrList Output;
    DCC::StatsP(Output);
    for (StrList::iterator pos = Output.begin(); pos != Output.end(); ++pos)
    {
      IRC::Notice(std::string(nick), *pos);
    }
#endif
  }
}

void List_Nicks(StrList & Output, const std::string & Nick)
{
  HASHREC *userptr;
  int i;    
  int numfound=0;
  char outmsg[MAX_BUFF];
  char fulluh[MAX_HOST+MAX_DOMAIN];

  Output.clear();

  for (i=0;i<HASHTABLESIZE;++i)
    {
      userptr = domaintable[i];
      while (userptr)
        {
          if (MatchesMask(userptr->info->nick, Nick))
            {
              if(!numfound++)
                { 
                  Output.push_back("The following clients match " + Nick + ":");
                }
              (void)sprintf(fulluh,
                            "%s@%s", userptr->info->user, userptr->info->host);
              (void)sprintf(outmsg,
                            "  %s (%s)", userptr->info->nick, fulluh);
              Output.push_back(outmsg);
            }
          userptr = userptr->collision;
        }
    } 
    
  if (numfound > 0)
    (void)sprintf(outmsg,
                  "%d matches for %s found",numfound, Nick.c_str());
  else    
    (void)sprintf(outmsg,
                  "No matches for %s found", Nick.c_str());
  Output.push_back(outmsg);
}

#ifdef AUTOPILOT
void SetAutoPilot(const std::string & From, const bool Val)
{
  autopilot = Val;
  if (Val) {
    ::Status("*** Autopilot turned ON by " + From, SEND_OPERS_ONLY);
    Log::Write("Autopilot turned ON by " + From);
  } else {
    ::Status("*** Autopilot turned OFF by " + From, SEND_OPERS_ONLY);
    Log::Write("Autopilot turned OFF by " + From);
  }
}
bool GetAutoPilot()
{
  return autopilot;
}
#endif

#ifdef SERVICES

typedef struct
{
  time_t last_checked_time;
  char   cloning_host[MAX_HOST];
  char   user_count[SMALL_BUFF];
  bool   kline_suggested;
} SERVICES_ENTRY;

SERVICES_ENTRY services;

void Check_Services()
{
  time_t cur_time;
  char clone_msg[MAX_BUFF];

  cur_time = time((time_t *) NULL);

  if ((services.last_checked_time + SERVICES_CHECK_TIME) < cur_time) {
    services.last_checked_time = cur_time;
#ifdef SERVICES
    (void)sprintf(clone_msg,"clones %d", SERVICES_CLONE_THRESHOLD);
    IRC::Msg(SERVICES_NICK, clone_msg);
#endif
#ifdef CA_SERVICES
    /* check if services.ca just signed on */
    IRC::IsOn("CA-SVCS");
#endif
  }
}

void On_Services_Notice(std::string Text)
{
  char text[MAX_BUFF];
  char *body;
  char *parm1;
  char *parm2;
  char *parm3;
  char dccbuff[MAX_BUFF];
  char userathost[MAX_HOST];
  bool identd;

  strncpy(text, Text.c_str(), sizeof(text));
  text[sizeof(text) - 1] = 0;
  body = text;

  while(*body == ' ')
    body++;

  parm1 = strtok(body," ");
  if(parm1 == (char *)NULL)
    return;

  parm2 = strtok((char *)NULL," ");
  if(parm2 == (char *)NULL)
    return;

  parm3 = strtok((char *)NULL,""); 
  if(parm3 == (char *)NULL)
    return;
  
  if( strstr(parm3,"users") )
    {
      strncpy(services.cloning_host, parm1, MAX_HOST-1);
      services.cloning_host[MAX_HOST - 1] = 0;
      strncpy(services.user_count, parm3, SMALL_BUFF-1);
      services.user_count[SMALL_BUFF - 1] = 0;
      services.kline_suggested = false;
      return;
    }
    
  if((strcasecmp("on", parm2) == 0) &&
    (strcasecmp(IRC::GetServerName().c_str(), parm3) == 0))
  {
    (void)sprintf(dccbuff,"%s reports %s cloning %s nick %s",
                    SERVICES_NAME,
                    services.user_count,
                    services.cloning_host,
                    parm1);

    ::Status(dccbuff, SEND_OPERS_ONLY);
    Log::Write(dccbuff);
  
    if(!services.kline_suggested)
    {
          char *user;
          char *host;
          char user_host[MAX_HOST+1];
      
          strncpy(userathost, services.cloning_host, MAX_HOST);
	  userathost[MAX_HOST - 1] = 0;
      
          /* strtok is going to destroy the original userathost,
             so save a copy for our own uses */
  
          strncpy(user_host, userathost, sizeof(user_host));
	  user_host[sizeof(user_host) - 1] = 0;
      
          user = strtok(userathost,"@");
          if(user == (char *)NULL)
            return; 
                    
          identd = true;
          if(*user == '~')
            identd = false;
  
          host = strtok((char *)NULL,"");
          if(host == (char *)NULL)
            return;  
          
          
#ifdef AUTOKSERVICES       /* If doing AUTO anything on any services */
#ifdef AUTO_KLINE_SERVICES /* If doing AUTO kline on any services */
          if((Config::IsOKHost(user_host)) || (Config::IsOper(user_host)))
	    return;
          
          Suggest_Kline(true,
                        user,
                        host,
                        false,
                        identd,
                        MK_CLONES);
            
#else                   /* doing AUTO on services.us, but killing only */
    
          if( (Config::IsOKHost(user_host)) && (Config::IsOper(user_host)))
	    return;
            
          if(autopilot)
            {
              IRC::Kill("Auto-Kill", parm1, "services.us clones");
            }
          
          Suggest_Kline(false,
                        user, 
                        host,
                        false,
                        identd,
                        MK_CLONES);
#endif                  
#else                   /* Not doing AUTO on services.us but reporting */
                        
          Suggest_Kline(false,
                        user,
                        host,
                        false,
                        identd,
                        MK_CLONES);
#endif    
          services.kline_suggested = true;
        }     
    }         
}

void On_CA_Services_Notice(std::string Text)
{
  char *text[MAX_BUFF];
  char *body;
  char *parm1, *parm2, *parm3, *parm4;
  char *user, *host, *tmp;
  char dccbuff[MAX_BUFF];
  char userathost[MAX_HOST];
  
        while(*body == ' ')
                body++;

        parm1 = strtok(body," ");
        if (parm1 == (char *)NULL)
                return;

        parm2 = strtok((char *)NULL," ");
        if(parm2 == (char *)NULL)
                return;

        parm3 = strtok((char *)NULL," ");
        if(parm3 == (char *)NULL)
                return;

        parm4 = strtok((char *)NULL," ");
        if(parm4 == (char *)NULL)
                return;
  
        /*
         * If this is services.ca suggesting a kline, we'll do it
         * ourselves, thanks very much.  Just return.
         */     
        if (strcasecmp("kline",parm2) == 0)
                return;
        
        if (strcasecmp("Nickflood",parm1) == 0)
          {
#ifdef AUTOKSERVICES
#ifdef AUTO_KILL_NICK_FLOODING
            if(autopilot)
              {
                IRC::Kill("Auto-Kill", parm3, "CaSvcs-Reported Nick Flooding");
                sprintf(dccbuff,
                        "AutoKilled %s %s for CaSvcs-Reported Nick Flooding",
                        parm3, parm4);
		::Status(dccbuff, SEND_OPERS_ONLY);
		Log::Write(dccbuff);
              }
            else
#endif   
#endif   
              { 
                sprintf(dccbuff,
                        "CaSvcs-Reported Nick flooding %s %s",
                        parm3, parm4);
		::Status(dccbuff, SEND_OPERS_ONLY);
		Log::Write(dccbuff);
              }

        }
            
        (void)sprintf(dccbuff,"%s: %s %s %s %s",
                CA_SERVICES_NAME, parm1, parm2, parm3, strtok(NULL, ""));
        ::Status(dccbuff, SEND_OPERS_ONLY); 
	Log::Write(dccbuff);
        
        /*              
         * Nope, services.ca is reporting clones.  Determine the
         * user@host.
         */ 

        tmp = strtok(dccbuff,"(");      /* Chop off up to the '(' */
        tmp = strtok((char *)NULL,")"); /* Chop off up to the ')' */
                
        user = strtok(tmp,"@");         /* Chop off up to the ')' */
        if (user == (char *)NULL)
                return;                 /* Hrm, strtok couldn't parse. */
              
        host = strtok((char *)NULL, "");
        if (host == (char *)NULL)
                return;                 /* Hrm, strtok couldn't parse. */

        if (strcasecmp("user-clones",parm2) == 0) {
                /*      
                 * Is this services.ca reporting multiple user@hosts?
                 */
                Suggest_Kline(false,user,host,false,(*user != '~'),MK_CLONES);
        } else if (strcasecmp("site-clones",parm2) == 0) {
                /*
                 * Multiple site clones?
                 */
                Suggest_Kline(false,user,host,true,(*user != '~'),MK_CLONES);
        } else if (strcasecmp("Nickflood",parm1) == 0) {
                /*
                 * Nick flooding twit.
                 */
                Suggest_Kline(false,user,host,false,(*user != '~'),MK_FLOODING);
        }
}

#endif

void CountClasses(StrList & Output, const std::string & classname)
{
  typedef std::map<std::string,int> ClassType;
  ClassType Classes;
  HASHREC *userptr;
  char outmsg[MAX_BUFF];
  int i;
  bool f;

  Output.clear();

  for (i = 0; i < HASHTABLESIZE; ++i)
  {
    userptr = domaintable[i];
    while (userptr)
    {
      Classes[userptr->info->userclass]++;
      userptr = userptr->collision;
    }
  }
  sprintf(outmsg, "%-10s %-6s %s", "Class", "Count", "Description");
  Output.push_back(outmsg);
  for (ClassType::iterator pos = Classes.begin(); pos != Classes.end();
    ++pos)
  {
    if ((classname == "") || (pos->first == classname))
    {
      sprintf(outmsg, "%-10s %-6d %s", pos->first.c_str(), pos->second,
	Config::GetYLineDescription(pos->first).c_str());
      Output.push_back(outmsg);
    }
  }
}

void Trace_Notice(std::string Text)
{
  char server_notice[MAX_BUFF];
  char *nick_reported;
  char *user_host;
  char *user;
  char *host;
  char *p;
  char notice[MAX_BUFF];
  char hold[MAX_BUFF];
  char copy_of_server_notice[MAX_BUFF];
  time_t current_time;
  int first_empty_entry = -1;
  bool found_entry = false;
  int i;

/* requested by anti__ (~anti@fax2.DCMETRONET.COM) [irc.blackened.com] */

  strncpy(server_notice, Text.c_str(), sizeof(server_notice));
  server_notice[sizeof(server_notice) - 1] = 0;

  current_time = time((time_t *)NULL);

  p = server_notice;

  strncpy(copy_of_server_notice, p, sizeof(copy_of_server_notice));
  copy_of_server_notice[sizeof(copy_of_server_notice) - 1] = 0;

  p = strtok(copy_of_server_notice," ");
  if(p == (char *)NULL) return; /* just ignore it *sigh* */
  
  if(strcasecmp(p,"requested") != 0)return;     /* just ignore it *sigh* */
  
  p = strtok((char *)NULL," ");
  if(p == (char *)NULL) return;         /* just ignore it *sigh* */
  if(strcasecmp(p,"by") != 0)return;    /* just ignore it *sigh* */
  
  nick_reported = strtok((char *)NULL," ");
  if(nick_reported == (char *)NULL)return;
      
  user_host = strtok((char *)NULL," ");
  if(user_host == (char *)NULL)return;  
      
/*
  Lets try and get it right folks... [user@host] or (user@host)
*/
  
  if (*user_host == '[') {
    user_host++;
    p = strrchr(user_host,']'); 
    if (p)
      *p = '\0'; 
  } else if (*user_host == '(') {
    user_host++;
    p = strrchr(user_host,')');
    if (p)
      *p = '\0';
  }
      
  /* Don't even complain about opers */
  
  if (Config::IsOper(user_host))
    return;
      
  ::Status("[TRACE " + Text + "]", SEND_OPERS_ONLY);
    
  for(i = 0; i < TRACE_TABLE_SIZE; i++ )
    {
      if(trace[i].user_host)
        { 
          if(strcasecmp(trace[i].user_host,user_host) == 0)
            {
              char *user;
              char *host;
  
              found_entry = true;
  
              /* if its an old old entry, let it drop to 0, then start counting
                 (this should be very unlikely case)
                 */
              if((trace[i].last_trace + MAX_TRACE_TIME) < current_time)
                {
                  trace[i].trace_count = 0;
                }
  
              trace[i].trace_count++;
      
              if(trace[i].trace_count >= MAX_TRACE_LOOKS)
                {
                  ::Status("Possible trace looker: " +
		    std::string(nick_reported) + " (" +
		    std::string(user_host) + ")", SEND_OPERS_ONLY);
		  Log::Write("Possible trace looker: " +
		    std::string(nick_reported) + " (" +
		    std::string(user_host) + ")");
  
/*              
   The code as it is, doesn't AUTO_KLINE link lookers ever, but will
   AUTO_KILL link lookers IFF autopilot is on. HOWEVER,
   Shadowfax still wants to see a suggested kline to use.
     
   - Dianora
*/            
                  if ( !Config::IsOKHost(user_host) )
                    {
#ifdef AUTO_KILL_TRACE_LOOKERS   
                      if(autopilot)
                        {
                          IRC::Kill("Auto-Kill", nick_reported, "repeated issuance of the /trace command");
                          (void)free(trace[i].user_host);
                          trace[i].user_host = (char *)NULL;
                        }     
#endif                        
                      user = strtok(user_host,"@");
                      if(user == (char *)NULL)return; /* OOOPSIES! */
                      host = strtok((char *)NULL,"");
                      if(host == (char *)NULL)return; /* OOPSIES! */
   
                      if(*user_host == '~') 
                        Suggest_Kline(false,user,host,false,false,MK_TRACE);
                      else
                        Suggest_Kline(false,user,host,false,true,MK_TRACE);
                    }
                    
#ifndef AUTO_KILL_TRACE_LOOKERS  
                  trace[i].last_trace = current_time;
#endif
                }       
              else            
                {     
                  trace[i].last_trace = current_time;
                }     
            }         
          else
            {         
              if((trace[i].last_trace + MAX_TRACE_TIME) < current_time)
                {     
                  (void)free(trace[i].user_host);
                  trace[i].user_host = (char *)NULL;
                }   
            }
        }         
      else
        {         
          if(first_empty_entry < 0)
            first_empty_entry = i;
        }             
    }               
                
/*            
   If this is a new entry, then found_entry will still be NO
*/                
                
  if(!found_entry)    
    {     
      if(first_empty_entry >= 0)
        {     
          trace[first_empty_entry].user_host = strdup(user_host);
          trace[first_empty_entry].last_trace = current_time;
          trace[first_empty_entry].trace_count = 0;
        }   
    }   
}

void MOTD_Notice(std::string Text)
{
  char server_notice[MAX_BUFF];
  char *nick_reported;
  char *user_host;
  char *user;
  char *host;
  char *p;
  char notice[MAX_BUFF];
  char hold[MAX_BUFF];
  char copy_of_server_notice[MAX_BUFF];
  time_t current_time;
  int first_empty_entry = -1;
  bool found_entry = false;
  int i;

/* requested by bubalu (~cray@Dial-1-85.funtv.com) [irc.blackened.com] */

  strncpy(server_notice, Text.c_str(), sizeof(server_notice));
  server_notice[sizeof(server_notice) - 1] = 0;

  current_time = time((time_t *)NULL);

  p = server_notice;

  strncpy(copy_of_server_notice, p, sizeof(copy_of_server_notice));
  copy_of_server_notice[sizeof(copy_of_server_notice) - 1] = 0;
  
  p = strtok(copy_of_server_notice," ");
  if(p == (char *)NULL) return; /* just ignore it *sigh* */
  
  if(strcasecmp(p,"requested") != 0)return;     /* just ignore it *sigh* */
  
  p = strtok((char *)NULL," ");
  if(p == (char *)NULL) return;         /* just ignore it *sigh* */
  if(strcasecmp(p,"by") != 0)return;    /* just ignore it *sigh* */
  
  nick_reported = strtok((char *)NULL," ");
  if(nick_reported == (char *)NULL)return;
      
  user_host = strtok((char *)NULL," ");
  if(user_host == (char *)NULL)return;  
      
/*
  Lets try and get it right folks... [user@host] or (user@host)
*/
  
  if (*user_host == '[') {
    user_host++;
    p = strrchr(user_host,']'); 
    if (p)
      *p = '\0'; 
  } else if (*user_host == '(') {
    user_host++;
    p = strrchr(user_host,')');
    if (p)
      *p = '\0';
  }
      
  /* Don't even complain about opers */
  
  if (Config::IsOper(user_host))
    return;
      
  ::Status("[MOTD " + Text + "]", SEND_OPERS_ONLY);
    
  for(i = 0; i < MOTD_TABLE_SIZE; i++ )
    {
      if(motd[i].user_host)
        { 
          if(strcasecmp(motd[i].user_host,user_host) == 0)
            {
              char *user;
              char *host;
  
              found_entry = true;
  
              /* if its an old old entry, let it drop to 0, then start counting
                 (this should be very unlikely case)
                 */
              if((motd[i].last_motd + MAX_MOTD_TIME) < current_time)
                {
                  motd[i].motd_count = 0;
                }
  
              motd[i].motd_count++;
      
              if(motd[i].motd_count >= MAX_MOTD_LOOKS)
                {
                  ::Status("Possible MOTD flooder: " +
		    std::string(nick_reported) + " (" +
		    std::string(user_host) + ")", SEND_OPERS_ONLY);
		  Log::Write("Possible MOTD flooder: " + std::string(nick_reported) +
		    " (" + std::string(user_host) + ")");
  
/*              
   The code as it is, doesn't AUTO_KLINE link lookers ever, but will
   AUTO_KILL link lookers IFF autopilot is on. HOWEVER,
   Shadowfax still wants to see a suggested kline to use.
     
   - Dianora
*/            
                  if ( !Config::IsOKHost(user_host) )
                    {
#ifdef AUTO_KILL_MOTD_LOOKERS   
                      if(autopilot)
                        {
                          IRC::Kill("Auto-Kill", nick_reported, "MOTD flooder");
                          (void)free(motd[i].user_host);
                          motd[i].user_host = (char *)NULL;
                        }     
#endif                        
                      user = strtok(user_host,"@");
                      if(user == (char *)NULL)return; /* OOOPSIES! */
                      host = strtok((char *)NULL,"");
                      if(host == (char *)NULL)return; /* OOPSIES! */
   
                      if(*user_host == '~') 
                        Suggest_Kline(false,user,host,false,false,MK_MOTD);
                      else
                        Suggest_Kline(false,user,host,false,true,MK_MOTD);
                    }
                    
#ifndef AUTO_KILL_MOTD_LOOKERS  
                  motd[i].last_motd = current_time;
#endif
                }       
              else            
                {     
                  motd[i].last_motd = current_time;
                }     
            }         
          else
            {         
              if((motd[i].last_motd + MAX_MOTD_TIME) < current_time)
                {     
                  (void)free(motd[i].user_host);
                  motd[i].user_host = (char *)NULL;
                }   
            }
        }         
      else
        {         
          if(first_empty_entry < 0)
            first_empty_entry = i;
        }             
    }               
                
/*            
   If this is a new entry, then found_entry will still be NO
*/                
                
  if(!found_entry)    
    {     
      if(first_empty_entry >= 0)
        {     
          motd[first_empty_entry].user_host = strdup(user_host);
          motd[first_empty_entry].last_motd = current_time;
          motd[first_empty_entry].motd_count = 0;
        }   
    }   
}

void Info_Notice(std::string Text)
{
  char server_notice[MAX_BUFF];
  char *nick_reported;
  char *user_host;
  char *user;
  char *host;
  char *p;
  char notice[MAX_BUFF];
  char hold[MAX_BUFF];
  char copy_of_server_notice[MAX_BUFF];
  time_t current_time;
  int first_empty_entry = -1;
  bool found_entry = false;
  int i;

/* requested by bubalu (~cray@Dial-1-85.funtv.com) [irc.blackened.com] */

  strncpy(server_notice, Text.c_str(), sizeof(server_notice));
  server_notice[sizeof(server_notice) - 1] = 0;

  current_time = time((time_t *)NULL);

  p = server_notice;

  strncpy(copy_of_server_notice, p, sizeof(copy_of_server_notice));
  copy_of_server_notice[sizeof(copy_of_server_notice) - 1] = 0;
  
  p = strtok(copy_of_server_notice," ");
  if(p == (char *)NULL) return; /* just ignore it *sigh* */
  
  if(strcasecmp(p,"requested") != 0)return;     /* just ignore it *sigh* */
  
  p = strtok((char *)NULL," ");
  if(p == (char *)NULL) return;         /* just ignore it *sigh* */
  if(strcasecmp(p,"by") != 0)return;    /* just ignore it *sigh* */
  
  nick_reported = strtok((char *)NULL," ");
  if(nick_reported == (char *)NULL)return;
      
  user_host = strtok((char *)NULL," ");
  if(user_host == (char *)NULL)return;  
      
/*
  Lets try and get it right folks... [user@host] or (user@host)
*/
  
  if (*user_host == '[') {
    user_host++;
    p = strrchr(user_host,']'); 
    if (p)
      *p = '\0'; 
  } else if (*user_host == '(') {
    user_host++;
    p = strrchr(user_host,')');
    if (p)
      *p = '\0';
  }
      
  /* Don't even complain about opers */
  
  if (Config::IsOper(user_host))
    return;
      
  ::Status("[INFO " + Text + "]", SEND_OPERS_ONLY);
    
  for(i = 0; i < INFO_TABLE_SIZE; i++ )
    {
      if(info[i].user_host)
        { 
          if(strcasecmp(info[i].user_host,user_host) == 0)
            {
              char *user;
              char *host;
  
              found_entry = true;
  
              /* if its an old old entry, let it drop to 0, then start counting
                 (this should be very unlikely case)
                 */
              if((info[i].last_info + MAX_INFO_TIME) < current_time)
                {
                  info[i].info_count = 0;
                }
  
              info[i].info_count++;
      
              if(info[i].info_count >= MAX_INFO_LOOKS)
                {
                  ::Status("Possible Info flooder: " +
		    std::string(nick_reported) + " (" +
		    std::string(user_host) + ")", SEND_OPERS_ONLY);
		  Log::Write("Possible Info flooder: " + std::string(nick_reported) +
		    " (" + std::string(user_host) + ")");
  
/*              
   The code as it is, doesn't AUTO_KLINE link lookers ever, but will
   AUTO_KILL link lookers IFF autopilot is on. HOWEVER,
   Shadowfax still wants to see a suggested kline to use.
     
   - Dianora
*/            
                  if ( !Config::IsOKHost(user_host) )
                    {
#ifdef AUTO_KILL_INFO_LOOKERS   
                      if(autopilot)
                        {
                          IRC::Kill("Auto-Kill", nick_reported, "Info flooder");
                          (void)free(info[i].user_host);
                          info[i].user_host = (char *)NULL;
                        }     
#endif                        
                      user = strtok(user_host,"@");
                      if(user == (char *)NULL)return; /* OOOPSIES! */
                      host = strtok((char *)NULL,"");
                      if(host == (char *)NULL)return; /* OOPSIES! */
   
                      if(*user_host == '~') 
                        Suggest_Kline(false,user,host,false,false,MK_INFO);
                      else
                        Suggest_Kline(false,user,host,false,true,MK_INFO);
                    }
                    
#ifndef AUTO_KILL_INFO_LOOKERS  
                  info[i].last_info = current_time;
#endif
                }       
              else            
                {     
                  info[i].last_info = current_time;
                }     
            }         
          else
            {         
              if((info[i].last_info + MAX_INFO_TIME) < current_time)
                {     
                  (void)free(info[i].user_host);
                  info[i].user_host = (char *)NULL;
                }   
            }
        }         
      else
        {         
          if(first_empty_entry < 0)
            first_empty_entry = i;
        }             
    }               
                
/*            
   If this is a new entry, then found_entry will still be NO
*/                
                
  if(!found_entry)    
    {     
      if(first_empty_entry >= 0)
        {     
          info[first_empty_entry].user_host = strdup(user_host);
          info[first_empty_entry].last_info = current_time;
          info[first_empty_entry].info_count = 0;
        }   
    }   
}

static void
Add_IP(std::string ip)
{
  unsigned long IP;
  time_t current_time;
  int first_empty_entry = -1;
  bool found_entry = false;
  int i, l;

  current_time = time((time_t *)NULL);

  IP = inet_addr(ip.c_str());

  ip = ip.substr(0, ip.rfind('.')) + ".";

  for(i = 0; i < IP_TABLE_SIZE; i++ )
    {
      if(ips[i].net)
        { 
          if(ips[i].net == (IP & vhost_netmask))
            {
              found_entry = true;
  
              /* if its an old old entry, let it drop to 0, then start counting
                 (this should be very unlikely case)
                 */
              if((ips[i].last_ip + MAX_IP_TIME) < current_time)
                {
                  ips[i].ip_count = 0;
                  ips[i].total_count = 0;
		  ips[i].first_ip = current_time;
                }
  
              ips[i].ip_count++;
	      ips[i].total_count++;
      
              if(ips[i].ip_count >= MAX_IP_CONNECTS)
                {
                  char turd[MAX_BUFF];
		  sprintf(turd, "Possible VHost clones from %s* (%d connects in %d seconds)", ip.c_str(), ips[i].total_count, current_time - ips[i].first_ip);
		  ::Status(turd, SEND_OPERS_ONLY);
		  Log::Write(turd);

                  ips[i].ip_count = 0;
                  ips[i].last_ip = current_time;
                }       
              else            
                {     
                  ips[i].last_ip = current_time;
                }     
            }         
          else
            {         
              if((ips[i].last_ip + MAX_IP_TIME) < current_time)
                {     
                  ips[i].net = 0;
                }   
            }
        }         
      else
        {         
          if(first_empty_entry < 0)
            first_empty_entry = i;
        }             
    }               
                
/*            
   If this is a new entry, then found_entry will still be NO
*/                
                
  if(!found_entry)    
    {     
      if(first_empty_entry >= 0)
        {     
          ips[first_empty_entry].net = IP & vhost_netmask;
          ips[first_empty_entry].last_ip = current_time;
          ips[first_empty_entry].first_ip = current_time;
          ips[first_empty_entry].ip_count = 1;
          ips[first_empty_entry].total_count = 1;
        }   
    }   
}

void List_IPs(StrList & Output, const std::string & IPMask)
{
  HASHREC *userptr;
  char fulluh[MAX_USER + MAX_DOMAIN + 2],
       fullip[MAX_USER + MAX_IPADDR + 2];
  char outmsg[MAX_BUFF];
  std::string IP;
  struct in_addr in;
  int i, numfound = 0;
  bool haveat;

  Output.clear();

  haveat = (IPMask.find('@') != std::string::npos);

  if (IPMask == "*" || IPMask == "*@*") {
    Output.push_back("Listing all users is not recommended. To do it anyway,");
    Output.push_back("use 'iplist ?*@*'.");
  } else {
    for (i = 0; i < HASHTABLESIZE; ++i) {
      userptr = iptable[i];
      while (userptr) {
	in.s_addr = userptr->info->ip;
	IP = inet_ntoa(in);
	sprintf(fulluh, "%s@%s", userptr->info->user,
	  userptr->info->host);
	sprintf(fullip, "%s@%s", userptr->info->user, IP.c_str());

	if (haveat ? MatchesMask(fullip, IPMask) : MatchesMask(IP, IPMask))
	{
	  if (!numfound++)
	    Output.push_back("The following clients match " + IPMask + ":");
	  sprintf(outmsg, "  %s (%s) [%s]", userptr->info->nick,
	    fulluh, IP.c_str());
	  Output.push_back(outmsg);
	}
	userptr = userptr->collision;
      }
    }
    if (numfound > 0)
      sprintf(outmsg, "%d matches for %s found", numfound,
        IPMask.c_str());
    else
      sprintf(outmsg, "No matches for %s found", IPMask.c_str());
    Output.push_back(outmsg);
  }
}

static	bool
CheckDNSBL(const std::string & ip, const std::string & host,
  const std::string & nick, const std::string & userhost)
{
  std::string::size_type nonNumeric = ip.find_first_not_of("0123456789.");
  int dots = CountChars(ip, '.');
  std::string dns = PROXY_BLACKHOLE_ZONE;

  if ((dots == 3) && (nonNumeric == std::string::npos))
  {
    std::string::size_type pos = 0;

    // Reverse the order of the octets and use them to prefix the blackhole
    // zone.
    while (dots >= 0)
    {
      std::string::size_type nextDot = ip.find('.', pos);
      dns = ip.substr(pos, nextDot - pos) + "." + dns;
      pos = nextDot + 1;
      dots--;
    }

    struct hostent *result = gethostbyname(dns.c_str());

    if (NULL != result)
    {
      // This is a proxy listed by the DNSBL
      std::string notice = std::string("DNSBL Open Proxy detected for ") +
	nick + "!" + userhost + " [" + ip + "]";
      Log::Write(notice);
      ::Status(notice, SEND_OPERS_ONLY);

      if (GetAutoPilot())
      {
#if defined(AUTO_DLINE_PROXIES)
        IRC::Dline("Auto-Dline", ip, REASON_DNSBL);
#elif defined(AUTO_KLINE_PROXIES)
# ifdef KLINE_IP_SUPPORTED
        IRC::Kline("Auto-Kline", DNSBL_KLINE_TIME, std::string("*@") + ip,
	  REASON_DNSBL);
# else
        IRC::Kline("Auto-Kline", DNSBL_KLINE_TIME, std::string("*@") + host,
	  REASON_DNSBL);
# endif
#endif
      }
      return true;
    }
  }
  return false;
}

void
CheckProxy(const std::string & ip, const std::string & host,
  const std::string & nick, const std::string & userhost)
{
  #ifdef PROXY_BLACKHOLE_ZONE
    // If the IP is listed by the DNSBL, there's no reason to do a Wingate
    // or SOCKS proxy check of our own!
    if (CheckDNSBL(ip, host, nick, userhost))
      return;
  #endif

  #ifdef DETECT_WINGATES
    Proxy::Connect(ip, Proxy::WINGATE, host, nick, userhost);
  #endif

  #ifdef DETECT_PROXIES
    Proxy::Connect(ip, Proxy::SOCKS4, host, nick, userhost);
  #endif
}

