// $Id: util.cc,v 1.33.2.1 2003/01/17 21:30:05 toast Exp $

// C headers
#include <sys/time.h>
#include <unistd.h>
#include <time.h>

// C++ headers
#include <string>

#include "strtype"
#include "oomon.h"
#include "util.h"
#include "dcc.h"
#include "remote.h"

// Originally I used strsep() to do this, but quickly learned how
// unportable it makes my code. *sigh* oh well. This works too.
int StrSplit(StrVector & temp, std::string input,
  const std::string & token)
{
  char		*item;
  int		count = 0;
  int		p, pos;

  temp.clear();
  while (input != "")
  {
    pos = input.length();
    for (int i = 0; i < token.length(); i++)
    {
      if ((p = input.find(token[i])) < pos)
      {
        pos = p;
      }
    }
    if ((pos < 0) || (pos >= input.length()))
    {
      temp.push_back(input);
      input = "";
      count++;
    }
    else
    {
      temp.push_back(input.substr(0, pos));
      input = input.substr(pos + 1);
      count++;
    }
  }
  return count;
}

int ReadLine(register int fd, std::string & text)
{
  int n, rc;
  char c;

  text = "";

  for (n = 1; n < MAX_BUFF; n++) {
    if ((rc = read(fd, &c, 1)) == 1) {
      if ((c != '\n') && (c != '\r'))
	text = text + c;
      else {
	//n--;
	break;
      }
    } else if (rc == 0) {
      if (n == 1)
	return 0;
      else
	break;
    } else
      return -1;
  }
  return n;
}

int SplitIRC(StrVector & temp, std::string text)
{
  int count = 0;

  temp.clear();
  while (text.length() > 0)
  {
    if (text.at(0) == ':')
    {
      text.erase((std::string::size_type) 0, 1);
      temp.push_back(text);
      text = std::string("");
      count++;
    }
    else
    {
      std::string::size_type i = text.find(' ');
      if (std::string::npos != i)
      {
	if (i == 0)
	{
	  text.erase((std::string::size_type) 0, 1);
	}
	else
	{
	  std::string chunk = text.substr(0, i);
          temp.push_back(chunk);
          text.erase(0, i + 1);
          count++;
	}
      }
      else
      {
        temp.push_back(text);
        text = "";
        count++;
      }
    }
  }
  return count;
}

void SplitFrom(std::string input, std::string & Nick, std::string & UserHost)
{
  int i = input.find('!');
  if (std::string::npos == i)
  {
    // This is a server, not a nick!user@host
    Nick = input;
    UserHost = "";
  }
  else
  {
    // This is a nick!user@host
    Nick = input.substr(0, i);
    UserHost = input.substr(i + 1);
  }
}

std::string
getNick(const std::string & text)
{
  std::string::size_type bang = text.find('!');

  if (std::string::npos == bang)
  {
    return text;
  }
  else
  {
    return text.substr(0, bang);
  }
}

int IRCCommand(const std::string & text)
{
  if (text == std::string("PING"))
    return IRC_PING;
  else if (text == std::string("NOTICE"))
    return IRC_NOTICE;
  else if (text == std::string("PRIVMSG"))
    return IRC_PRIVMSG;
  else if (text == std::string("WALLOPS"))
    return IRC_WALLOPS;
  else
    return IRC_UNKNOWN;
}

unsigned long atoul(const char *text)
{
  unsigned long temp = 0;
  int len = strlen(text);
  while ((*text >= '0') && (*text <= '9'))
    temp = temp * 10 + (*(text++) - '0');
  return temp;
}

int DCCCommand(const std::string & text)
{
  if (text == "AUTH") return DCC_AUTH;
  else if (text == "QUIT") return DCC_QUIT;
  else if (text == "CHAT") return DCC_CHAT;
  else if (text == "WHO") return DCC_WHO;
  else if (text == "LINKS") return DCC_LINKS;
  else if (text == "HELP") return DCC_HELP;
  else if (text == "WATCH") return DCC_WATCH;
  else if (text == "MOTD") return DCC_MOTD;
  else if (text == "KLINE") return DCC_KLINE;
  else if (text == "UNKLINE") return DCC_UNKLINE;
  else if (text == "KCLONE") return DCC_KCLONE;
  else if (text == "KFLOOD") return DCC_KFLOOD;
  else if (text == "KBOT") return DCC_KBOT;
  else if (text == "KLINK") return DCC_KLINK;
  else if (text == "KSPAM") return DCC_KSPAM;
  else if (text == "KPERM") return DCC_KPERM;
  else if (text == "GLINE") return DCC_GLINE;
  else if (text == "UNGLINE") return DCC_UNGLINE;
  else if (text == "GUNKLINE") return DCC_UNGLINE;
  else if (text == "GCLONE") return DCC_GCLONE;
  else if (text == "GFLOOD") return DCC_GFLOOD;
  else if (text == "GBOT") return DCC_GBOT;
  else if (text == "GLINK") return DCC_GLINK;
  else if (text == "GSPAM") return DCC_GSPAM;
  else if (text == "GPERM") return DCC_GPERM;
  else if (text == "DLINE") return DCC_DLINE;
  else if (text == "KILL") return DCC_KILL;
  else if (text == "KILLLIST") return DCC_KILLLIST;
  else if (text == "KL") return DCC_KILLLIST;
  else if (text == "MULTI") return DCC_MULTI;
  else if (text == "HMULTI") return DCC_HMULTI;
  else if (text == "UMULTI") return DCC_UMULTI;
  else if (text == "VMULTI") return DCC_VMULTI;
  else if (text == "BOTS") return DCC_MULTI;
  else if (text == "VBOTS") return DCC_VMULTI;
  else if (text == "NFIND") return DCC_NFIND;
  else if (text == "TRACE") return DCC_TRACE;
  else if (text == "CLONES") return DCC_CLONES;
  else if (text == "LIST") return DCC_LIST;
  else if (text == "IPLIST") return DCC_IPLIST;
  else if (text == "FINDK") return DCC_FINDK;
  else if (text == "RELOAD") return DCC_RELOAD;
  else if (text == "DOMAINS") return DCC_DOMAINS;
  else if (text == "CLASS") return DCC_CLASS;
  else if (text == "DIE") return DCC_DIE;
  else if (text == "CONN") return DCC_CONN;
  else if (text == "DISCONN") return DCC_DISCONN;
  else if (text == "AUTOPILOT") return DCC_AUTOPILOT;
  else if (text == "KTRACE") return DCC_KTRACE;
  else if (text == "KMOTD") return DCC_KMOTD;
  else if (text == "GTRACE") return DCC_GTRACE;
  else if (text == "GMOTD") return DCC_GMOTD;
  else if (text == "KINFO") return DCC_KINFO;
  else if (text == "GINFO") return DCC_GINFO;
  else if (text == "KWINGATE") return DCC_KWINGATE;
  else if (text == "GWINGATE") return DCC_GWINGATE;
  else if (text == "KPROXY") return DCC_KPROXY;
  else if (text == "GPROXY") return DCC_GPROXY;
  else if (text == "RAW") return DCC_RAW;
  else if (text == "OP") return DCC_OP;
  else if (text == "JOIN") return DCC_JOIN;
  else if (text == "PART") return DCC_PART;
  else if (text == "TRAP") return DCC_TRAP;
  else if (text == "UNDLINE") return DCC_UNDLINE;
  else if (text == "FINDD") return DCC_FINDD;
  else if (text == "ECHO") return DCC_ECHO;
#ifdef DEBUG
  else if (text == "TEST") return DCC_TEST;
#endif
  else return DCC_UNKNOWN;
}

std::string FirstWord(std::string & Text)
{
  std::string temp;
  int i = Text.find(' ');
  if (std::string::npos != i)
  {
    temp = Text.substr(0, i);
    Text = Text.substr(i + 1);
  }
  else
  {
    temp = Text;
    Text = "";
  }
  while ((Text.length() > 0) && (Text[0] == ' '))
    Text.erase((std::string::size_type) 0, 1);
  return temp;
}

std::string UpCase(const std::string & Text)
{
  std::string result = Text;
  std::string::size_type len = Text.length();

  for (std::string::size_type index = 0; index < len; index++)
  {
	if ((result[index] >= 'a') && (result[index] <= 'z'))
	{
		result[index] = (result[index] - 'a' + 'A');
	}
  }

  return result;
}

std::string DownCase(const std::string & Text)
{
  std::string result = Text;
  std::string::size_type len = Text.length();

  for (std::string::size_type index = 0; index < len; index++)
  {
	if ((result[index] >= 'A') && (result[index] <= 'Z'))
	{
		result[index] = (result[index] - 'A' + 'a');
	}
  }

  return result;
}

bool Same(std::string Text1, std::string Text2)
{
  return (UpCase(Text1) == UpCase(Text2));
}

bool FixedMatch(std::string TEST, std::string MASK)
{
  if (TEST == MASK)
  {
    return true;
  }
  else
  {
    if (TEST.length() == MASK.length())
    {
      for (int i = 0; i < MASK.length(); i++)
      {
	if ((MASK[i] != '?') && (MASK[i] != TEST[i]))
	{
	  return false;
	}
      }
    }
    else
    {
      return false;
    }
  }
  return true;
}

bool SMatch(std::string TEST, std::string MASK)
{
  if ((MASK == "*") || (TEST == MASK))
    return true;
  if ((MASK == "") || (TEST == ""))
    return false;
  if ((MASK[0] == '*') && (std::string::npos != MASK.substr(1).find('*')))
  {
    MASK.erase((std::string::size_type) 0, 1);
    for (int i = 0; i < TEST.length(); i++)
    {
      std::string temps = TEST.substr(i);
      int p = MASK.find('*');
      if ((std::string::npos != p) &&
	FixedMatch(temps.substr(0, p), MASK.substr(0, p)))
      {
	temps = temps.substr(p);
	std::string tempm = MASK.substr(p);
	if (SMatch(temps, tempm))
	  return true;
      }
    }
    return false;
  }
  else if (MASK[0] == '*')
  {
    MASK.erase((std::string::size_type) 0, 1);
    TEST.erase(0, TEST.length() - MASK.length());
    return FixedMatch(TEST, MASK);
  }
  else
  {
    int p = MASK.find('*');
    if (std::string::npos != p)
    {
      std::string temps = TEST.substr(0, p);
      std::string tempm = MASK.substr(0, p);
      if (FixedMatch(temps, tempm))
      {
        TEST.erase(0, p);
        MASK.erase(0, p);
        return SMatch(TEST, MASK);
      }
      else
      {
        return false;
      }
    }
    else
    {
      return FixedMatch(TEST, MASK);
    }
  }
}

bool MatchesMask(std::string TEST, std::string MASK)
{
  return SMatch(UpCase(TEST), UpCase(MASK));
}

bool ChkPass(std::string CORRECT, std::string TEST)
{
  if (CORRECT == "*")
    return false;
  if ((CORRECT == "") && (TEST == ""))
    return true;
#ifdef USE_CRYPT
  if (CORRECT == crypt(TEST.c_str(), CORRECT.c_str()))
    return true;
#else
  return (CORRECT == TEST);
#endif
  return false;
}

std::string TimeStamp()
{
  char ts_holder[MAX_BUFF];
  struct tm *time_val;
  time_t t;
  
  t = time(0);
  time_val = localtime(&t);

  strftime(ts_holder, MAX_BUFF-1, "%c", time_val);

  return std::string(ts_holder);
}

std::string IntToStr(int i)
{
  if (i < 0)
    return std::string("-") + IntToStr(-i);
  else if (i == 0)
    return std::string("0");
  else {
    std::string inttostr_temp = std::string("");
    while (i > 0) {
      inttostr_temp.insert((std::string::size_type) 0,
        (std::string::size_type) 1, (std::string::value_type)((i % 10) + '0'));
      i /= 10;
    }
    return inttostr_temp;
  }
}

std::string ULongToStr(unsigned long i)
{
  if (i == 0)
    return std::string("0");
  else {
    std::string ulongtostr_temp = std::string("");
    while (i != 0) {
      ulongtostr_temp.insert((std::string::size_type) 0,
        (std::string::size_type) 1, (std::string::value_type)((i % 10) + '0'));
      i /= 10;
    }
    return ulongtostr_temp;
  }
}

int RemoteCommand(const std::string & text)
{
  if (text == "ERROR") return RC_ERROR;
  else if (text == REM_STATUS) return RC_STATUS;
  else if (text == REM_CHAT) return RC_CHAT;
  else if (text == REM_WHO) return RC_WHO;
  else if (text == REM_LINKS) return RC_LINKS;
  else if (text == REM_KLINE) return RC_KLINE;
  else if (text == REM_KCLONE) return RC_KCLONE;
  else if (text == REM_KFLOOD) return RC_KFLOOD;
  else if (text == REM_KBOT) return RC_KBOT;
  else if (text == REM_KSPAM) return RC_KSPAM;
  else if (text == REM_KLINK) return RC_KLINK;
  else if (text == REM_KTRACE) return RC_KTRACE;
  else if (text == REM_KMOTD) return RC_KMOTD;
  else if (text == REM_KINFO) return RC_KINFO;
  else if (text == REM_KWINGATE) return RC_KWINGATE;
  else if (text == REM_KPROXY) return RC_KPROXY;
  else if (text == REM_KLINE) return RC_KPERM;
  else if (text == REM_UNKLINE) return RC_UNKLINE;
  else if (text == REM_GLINE) return RC_GLINE;
  else if (text == REM_GCLONE) return RC_GCLONE;
  else if (text == REM_GFLOOD) return RC_GFLOOD;
  else if (text == REM_GBOT) return RC_GBOT;
  else if (text == REM_GSPAM) return RC_GSPAM;
  else if (text == REM_GTRACE) return RC_GTRACE;
  else if (text == REM_GMOTD) return RC_GMOTD;
  else if (text == REM_GINFO) return RC_GINFO;
  else if (text == REM_GWINGATE) return RC_GWINGATE;
  else if (text == REM_GPROXY) return RC_GPROXY;
  else if (text == REM_GLINK) return RC_GLINK;
  else if (text == REM_GPERM) return RC_GPERM;
  else if (text == REM_UNGLINE) return RC_UNGLINE;
  else if (text == REM_CONNECT) return RC_CONNECT;
  else if (text == REM_DISCONN) return RC_DISCONN;
  else if (text == REM_DIE) return RC_DIE;
  else if (text == REM_MOTD) return RC_MOTD;
  return RC_UNKNOWN;
}

std::string GetBotHB(std::string text)
{
  if (text == "*")
    return text;

  int i;
  if ((i = text.find('@')) != std::string::npos)
    return text.substr(i + 1);
  else
    return "";
}

std::string GetHandleHB(std::string text)
{
  if (text == "*")
    return text;

  int i;
  if ((i = text.find('@')) != std::string::npos)
    return text.substr(0, i);
  else
    return "";
}

std::string::size_type CountChars(const std::string & text,
  const std::string::value_type ch)
{
  std::string::size_type length = text.length();
  std::string::size_type count = 0;

  for (std::string::size_type index = 0; index < length; index++)
  {
    if (text[index] == ch)
    {
      count++;
    }
  }

  return count;
}

std::string klineMask(const std::string & userhost)
{
  std::string::size_type at = userhost.find('@');
  if (at != std::string::npos)
  {
    std::string user = userhost.substr(0, at);
    std::string host = userhost.substr(at + 1, std::string::npos);

    if ((user.length() > 0) && (user[0] == '~'))
    {
      user = "~*";
    }
    else
    {
      user = "*" + user;
    }

    if ((std::string::npos == host.find_first_not_of(".0123456789")) &&
      (3 == CountChars(host, '.')))
    {
      std::string::size_type lastDot = host.rfind('.');
      return (user + "@" + host.substr(0, lastDot) + ".*");
    }
    else
    {
      std::string::size_type lastDot = host.rfind('.');
      if (std::string::npos == lastDot)
      {
	return (user + "@" + host);
      }
      else
      {
	std::string::size_type prevDot = host.rfind('.', lastDot - 1);
	std::string tld = host.substr(lastDot + 1, std::string::npos);
	if (std::string::npos == prevDot)
	{
	  // user@domain.tld
	  return (user + "@*" + host);
	}
	else
	{
	  // user@*.domain.tld
	  if (tld.length() == 2)
	  {
	    // user@*.domain.xx
	    std::string::size_type triDot = host.rfind('.', prevDot - 1);
	    if (std::string::npos == triDot)
	    {
	      // user@foo.domain.xx
	      if (std::string::npos == host.substr(0,
		prevDot).find_first_of("0123456789"))
              {
		// user@alpha.domain.xx
	        return (user + "@*" + host);
	      }
	      else
	      {
		// user@al55pha.domain.xx
	        return (user + "@*" + host.substr(prevDot, std::string::npos));
	      }
	    }
	    else
	    {
	      // user@*.foo.domain.xx
	      if (std::string::npos == host.substr(triDot + 1,
		prevDot - triDot - 1).find_first_of("0123456789"))
              {
		// user@*.alpha.domain.xx
	        return (user + "@*" + host.substr(triDot, std::string::npos));
	      }
	      else
	      {
		// user@*.al55pha.domain.xx
	        return (user + "@*" + host.substr(prevDot, std::string::npos));
	      }
	    }
	  }
	  else
	  {
	    // user@*.domain.tld
	    return (user + "@*" + host.substr(prevDot, std::string::npos));
	  }
	}
      }
    }
  }
  else
  {
    return userhost;
  }
}

std::string classCMask(const std::string & ip)
{
  std::string::size_type lastDot = ip.rfind('.');

  return (ip.substr(0, lastDot) + ".*");
}

